#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.database`."""

import pytest

from orion.core.io.database import Database, ReadOnlyDB
from orion.core.io.database.mongodb import MongoDB
from orion.core.utils.singleton import SingletonAlreadyInstantiatedError, \
    SingletonNotInstantiatedError


@pytest.mark.usefixtures("null_db_instances")
class TestDatabaseFactory(object):
    """Test the creation of a determinate `Database` type, by a complete spefication
    of a database by-itself (this on which every `Database` acts on as part
    of its being, attributes of an `AbstractDB`) and for-itself (what essentially
    differentiates one concrete `Database` from one other).

    """

    def test_empty_first_call(self):
        """Should not be able to make first call without any arguments.

        Hegelian Ontology Primer
        ------------------------

        Type indeterminate <-> type abstracted from its property <-> No type
        """
        with pytest.raises(SingletonNotInstantiatedError):
            Database()

    def test_notfound_type_first_call(self):
        """Raise when supplying not implemented wrapper name."""
        with pytest.raises(NotImplementedError) as exc_info:
            Database('notfound')

        assert 'AbstractDB' in str(exc_info.value)

    def test_instantiation_and_singleton(self):
        """Test create just one object, that object persists between calls."""
        database = Database(of_type='MongoDB', name='orion_test',
                            username='user', password='pass')

        assert isinstance(database, MongoDB)
        assert database is MongoDB()
        assert database is Database()

        with pytest.raises(SingletonAlreadyInstantiatedError):
            Database('fire', [], {'it_matters': 'it\'s singleton'})


@pytest.mark.usefixtures("null_db_instances", "clean_db")
class TestReadOnlyDatabase(object):
    """Test coherence of read-only database and its wrapped database."""

    def test_valid_attributes(self, create_db_instance):
        """Test attributes are coherent from view and wrapped database."""
        database = create_db_instance
        readonly_database = ReadOnlyDB(database)

        assert readonly_database.is_connected == database.is_connected
        assert readonly_database.host == database.host
        assert readonly_database.port == database.port
        assert readonly_database.username == database.username
        assert readonly_database.password == database.password

    def test_read(self, create_db_instance):
        """Test read is coherent from view and wrapped database."""
        database = create_db_instance
        readonly_database = ReadOnlyDB(database)

        args = {"collection_name": "trials", "query": {"experiment": "supernaedo2-dendi"}}
        readonly_result = readonly_database.read(**args)
        result = database.read(**args)

        assert len(result) > 0  # Otherwise the test is pointless
        assert readonly_result == result

    def test_invalid_attributes(self, create_db_instance):
        """Test that attributes for writing are not accessible."""
        database = create_db_instance
        readonly_database = ReadOnlyDB(database)

        # Test that database.ensure_index indeed exists
        database.ensure_index
        with pytest.raises(AttributeError):
            readonly_database.ensure_index

        # Test that database.write indeed exists
        database.write
        with pytest.raises(AttributeError):
            readonly_database.write

        # Test that database.read_and_write indeed exists
        database.read_and_write
        with pytest.raises(AttributeError):
            readonly_database.read_and_write

        # Test that database.remove indeed exists
        database.remove
        with pytest.raises(AttributeError):
            readonly_database.remove



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker.consumer` -- Evaluate objective on a set of parameters
==============================================================================

.. module:: consumer
   :platform: Unix
   :synopsis: Call user's script as a black box process to evaluate a trial.

"""
import copy
import logging
import os
import signal
import subprocess
import tempfile

import orion.core
from orion.core.io.orion_cmdline_parser import OrionCmdlineParser
from orion.core.io.resolve_config import infer_versioning_metadata
from orion.core.utils.exceptions import BranchingEvent, InexecutableUserScript
from orion.core.utils.working_dir import WorkingDir
from orion.core.worker.trial_pacemaker import TrialPacemaker

log = logging.getLogger(__name__)


# pylint: disable = unused-argument
def _handler(signum, frame):
    log.error('Oríon has been interrupted.')
    raise KeyboardInterrupt


class ExecutionError(Exception):
    """Error raised when Orion is unable to execute the user's script without errors."""

    pass


class Consumer(object):
    """Consume a trial by using it to initialize a black-box box to evaluate it.

    It uses an `Experiment` object to push an evaluated trial, if results are
    delivered to the worker process successfully.

    It forks another process which executes user's script with the suggested
    options. It expects results to be written in a **JSON** file, whose path
    has been defined in a special orion environmental variable which is set
    into the child process' environment.

    Parameters
    ----------
    experiment: `orion.core.worker.experiment.Experiment`
        Manager of this experiment, provides convenient interface for interacting with
        the database.

    heartbeat: int, optional
        Frequency (seconds) at which the heartbeat of the trial is updated.
        If the heartbeat of a `reserved` trial is larger than twice the configured
        heartbeat, Oríon will reset the status of the trial to `interrupted`.
        This allows restoring lost trials (ex: due to killed worker).
        Defaults to `orion.core.config.worker.heartbeat`.

    user_script_config: str, optional
        Config argument name of user's script (--config).
        Defaults to `orion.core.config.worker.user_script_config`.

    interrupt_signal_code: int, optional
        Signal returned by user script to signal to Oríon that it was interrupted.
        Defaults to `orion.core.config.worker.interrupt_signal_code`.

    """

    def __init__(self, experiment, heartbeat=None, user_script_config=None,
                 interrupt_signal_code=None):
        log.debug("Creating Consumer object.")
        self.experiment = experiment
        self.space = experiment.space
        if self.space is None:
            raise RuntimeError("Experiment object provided to Consumer has not yet completed"
                               " initialization.")

        if heartbeat is None:
            heartbeat = orion.core.config.worker.heartbeat

        if user_script_config is None:
            user_script_config = orion.core.config.worker.user_script_config

        if interrupt_signal_code is None:
            interrupt_signal_code = orion.core.config.worker.interrupt_signal_code

        self.heartbeat = heartbeat
        self.interrupt_signal_code = interrupt_signal_code

        # Fetch space builder
        self.template_builder = OrionCmdlineParser(user_script_config)
        self.template_builder.set_state_dict(experiment.metadata['parser'])
        # Get path to user's script and infer trial configuration directory
        if experiment.working_dir:
            self.working_dir = os.path.abspath(experiment.working_dir)
        else:
            self.working_dir = os.path.join(tempfile.gettempdir(), 'orion')

        self.pacemaker = None

    def consume(self, trial):
        """Execute user's script as a block box using the options contained
        within `trial`.

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            Orion trial to execute.

        Returns
        -------
        bool
            True if the trial was successfully executed. False if the trial is broken.

        """
        log.debug("### Create new directory at '%s':", self.working_dir)
        temp_dir = not bool(self.experiment.working_dir)
        prefix = self.experiment.name + "_"
        suffix = trial.id

        try:
            with WorkingDir(self.working_dir, temp_dir,
                            prefix=prefix, suffix=suffix) as workdirname:
                log.debug("## New consumer context: %s", workdirname)
                trial.working_dir = workdirname

                results_file = self._consume(trial, workdirname)

                log.debug("## Parse results from file and fill corresponding Trial object.")
                self.experiment.update_completed_trial(trial, results_file)

            success = True

        except KeyboardInterrupt:
            log.debug("### Save %s as interrupted.", trial)
            self.experiment.set_trial_status(trial, status='interrupted')
            raise

        except ExecutionError:
            log.debug("### Save %s as broken.", trial)
            self.experiment.set_trial_status(trial, status='broken')

            success = False

        return success

    def get_execution_environment(self, trial, results_file='results.log'):
        """Set a few environment variables to allow users and
        underlying processes to know if they are running under orion.

        Parameters
        ----------
        results_file: str
           file used to store results, this is only used by the legacy protocol

        trial: Trial
           reference to the trial object that is going to be run

        Notes
        -----
        This function defines the environment variables described below

        .. envvar:: ORION_EXPERIMENT_ID

           Current experiment that is being ran.

        .. envvar::  ORION_EXPERIMENT_NAME

           Name of the experiment the worker is currently working on.

        .. envvar::  ORION_EXPERIMENT_VERSION

           Version of the experiment the worker is currently working on.

        .. envvar:: ORION_TRIAL_ID

           Current trial id that is currently being executed in this process.

        .. envvar:: ORION_WORKING_DIRECTORY

           Trial's current working directory.

        .. envvar:: ORION_RESULTS_PATH

           Trial's results file that is read by the legacy protocol to get the results of the trial
           after a successful run.

        """
        env = dict(os.environ)
        env['ORION_EXPERIMENT_ID'] = str(self.experiment.id)
        env['ORION_EXPERIMENT_NAME'] = str(self.experiment.name)
        env['ORION_EXPERIMENT_VERSION'] = str(self.experiment.version)
        env['ORION_TRIAL_ID'] = str(trial.id)

        env['ORION_WORKING_DIR'] = str(trial.working_dir)
        env['ORION_RESULTS_PATH'] = str(results_file)
        env['ORION_INTERRUPT_CODE'] = str(self.interrupt_signal_code)

        return env

    def _consume(self, trial, workdirname):
        config_file = tempfile.NamedTemporaryFile(mode='w', prefix='trial_',
                                                  suffix='.conf', dir=workdirname,
                                                  delete=False)
        config_file.close()
        log.debug("## New temp config file: %s", config_file.name)
        results_file = tempfile.NamedTemporaryFile(mode='w', prefix='results_',
                                                   suffix='.log', dir=workdirname,
                                                   delete=False)
        results_file.close()
        log.debug("## New temp results file: %s", results_file.name)

        log.debug("## Building command line argument and configuration for trial.")
        env = self.get_execution_environment(trial, results_file.name)
        cmd_args = self.template_builder.format(config_file.name, trial, self.experiment)

        log.debug("## Launch user's script as a subprocess and wait for finish.")

        self._validate_code_version()

        self.pacemaker = TrialPacemaker(trial, self.heartbeat)
        self.pacemaker.start()
        try:
            self.execute_process(cmd_args, env)
        finally:
            # merciless
            self.pacemaker.stop()

        return results_file

    def _validate_code_version(self):
        old_config = self.experiment.configuration
        new_config = copy.deepcopy(old_config)
        new_config['metadata']['VCS'] = infer_versioning_metadata(
            old_config['metadata']['user_script'])

        # Circular import
        from orion.core.evc.conflicts import CodeConflict
        conflicts = list(CodeConflict.detect(old_config, new_config))
        if conflicts:
            raise BranchingEvent(f'Code changed between execution of 2 trials:\n{conflicts[0]}')

    # pylint: disable = no-self-use
    def execute_process(self, cmd_args, environ):
        """Facilitate launching a black-box trial."""
        command = cmd_args

        signal.signal(signal.SIGTERM, _handler)

        try:
            process = subprocess.Popen(command, env=environ)
        except PermissionError:
            log.debug("### Script is not executable")
            raise InexecutableUserScript(' '.join(cmd_args))

        return_code = process.wait()

        if return_code == self.interrupt_signal_code:
            raise KeyboardInterrupt()
        elif return_code != 0:
            raise ExecutionError("Something went wrong. Check logs. Process "
                                 "returned with code {} !".format(return_code))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform unit-test on functions user for insert cli."""
import numbers

import pytest

from orion.core.cli.insert import _validate_input_value
from orion.core.io.space_builder import SpaceBuilder


@pytest.fixture()
def real_space():
    """Fixture for real space"""
    return SpaceBuilder().build({"x": "uniform(-10,20)"})


@pytest.fixture()
def integer_space():
    """Fixture for integer space"""
    return SpaceBuilder().build({"x": "uniform(-10,20,discrete=True)"})


@pytest.fixture()
def categorical_space():
    """Fixture for categorical space"""
    return SpaceBuilder().build({"x": "choices([10.1,11,'12','string'])"})


def test_validate_input_value_real_real(real_space):
    """Test if real value passed to real space is validated properly"""
    namespace = 'x'
    is_valid, casted_value = _validate_input_value("10.0", real_space, namespace)
    assert is_valid
    assert isinstance(casted_value, numbers.Number)


def test_validate_input_value_real_integer(real_space):
    """Test if integer value passed to real space is validated properly"""
    namespace = 'x'
    is_valid, casted_value = _validate_input_value("10", real_space, namespace)
    assert is_valid
    assert isinstance(casted_value, numbers.Number)


def test_validate_input_value_real_string(real_space):
    """Test if string value passed to real space is rejected properly"""
    namespace = 'x'
    is_valid, casted_value = _validate_input_value("string", real_space, namespace)
    assert not is_valid


def test_validate_input_value_real_out_of_bound(real_space):
    """Test if out of bound values passed to real space are rejected properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("100.0", real_space, namespace)
    assert not is_valid

    is_valid, casted_value = _validate_input_value("100", real_space, namespace)
    assert not is_valid


def test_validate_input_value_integer_real(integer_space):
    """Test if real value passed to integer space is validated properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("10.0", integer_space, namespace)
    assert is_valid
    assert isinstance(casted_value, numbers.Number)


def test_validate_input_value_integer_integer(integer_space):
    """Test if integer value passed to integer space is validated properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("10", integer_space, namespace)
    assert is_valid
    assert isinstance(casted_value, numbers.Number)


def test_validate_input_value_integer_string(integer_space):
    """Test if string value passed to integer space is rejected properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("string", integer_space, namespace)
    assert not is_valid


def test_validate_input_value_integer_out_of_bound(integer_space):
    """Test if out of bound values passed to integer space are rejected properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("100.0", integer_space, namespace)
    assert not is_valid

    is_valid, casted_value = _validate_input_value("100", integer_space, namespace)
    assert not is_valid


def test_validate_input_value_categorical_real_hit(categorical_space):
    """Test if real value passed to categorical space is validated properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("10.1", categorical_space, namespace)
    assert is_valid
    assert isinstance(casted_value, numbers.Number)


def test_validate_input_value_categorical_real_nohit(categorical_space):
    """Test if bad real value passed to categorical space is rejected properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("10", categorical_space, namespace)
    assert not is_valid

    is_valid, casted_value = _validate_input_value("10.0", categorical_space, namespace)
    assert not is_valid

    is_valid, casted_value = _validate_input_value("10.2", categorical_space, namespace)
    assert not is_valid


def test_validate_input_value_categorical_integer_hit(categorical_space):
    """Test if integer value passed to categorical space is validated properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("11", categorical_space, namespace)
    assert is_valid
    assert isinstance(casted_value, numbers.Number)

    is_valid, casted_value = _validate_input_value("11.0", categorical_space, namespace)
    assert is_valid
    assert isinstance(casted_value, numbers.Number)


def test_validate_input_value_categorical_integer_nohit(categorical_space):
    """Test if bad integer value passed to categorical space is rejected properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("15", categorical_space, namespace)
    assert not is_valid


def test_validate_input_value_categorical_string_number(categorical_space):
    """Test if string number value passed to categorical space is validated properly"""
    namespace = 'x'

    # Make sure integer 12 does not pass
    is_valid, casted_value = _validate_input_value("12", categorical_space, namespace)
    assert not is_valid

    # Now test "12" as a string
    is_valid, casted_value = _validate_input_value("'12'", categorical_space, namespace)
    assert is_valid
    assert isinstance(casted_value, str)


def test_validate_input_value_categorical_string_value(categorical_space):
    """Test if literal string value passed to categorical space is validated properly"""
    namespace = 'x'

    is_valid, casted_value = _validate_input_value("random", categorical_space, namespace)
    assert not is_valid

    is_valid, casted_value = _validate_input_value("string", categorical_space, namespace)
    assert is_valid
    assert isinstance(casted_value, str)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the list command."""
import os

import orion.core.cli


def test_no_exp(monkeypatch, clean_db, capsys):
    """Test that nothing is printed when there are no experiments."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == "No experiment found\n"


def test_single_exp(clean_db, one_experiment, capsys):
    """Test that the name of the experiment is printed when there is one experiment."""
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-v1\n"


def test_no_version_backward_compatible(clean_db, one_experiment_no_version, capsys):
    """Test status with no experiments."""
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-no-version-v1\n"


def test_broken_refers(clean_db, broken_refers, capsys):
    """Test that experiment without refers dict can be handled properly."""
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-v1\n"


def test_python_api(clean_db, with_experiment_using_python_api, capsys):
    """Test list if containing exps from cmdline api and python api"""
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-v1\n from-python-api-v1\n"


def test_missing_conf_file(clean_db, with_experiment_missing_conf_file, capsys):
    """Test list can handle experiments when the user script config file is missing"""
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-v1\n"


def test_two_exp(capsys, clean_db, two_experiments):
    """Test that experiment and child are printed."""
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == """\
 test_double_exp-v1┐
                   └test_double_exp_child-v1
"""


def test_three_exp(capsys, clean_db, three_experiments):
    """Test that experiment, child  and grand-child are printed."""
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == """\
 test_double_exp-v1┐
                   └test_double_exp_child-v1
 test_single_exp-v1
"""


def test_no_exp_name(clean_db, three_experiments, monkeypatch, capsys):
    """Test that nothing is printed when there are no experiments with a given name."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list', '--name', 'I don\'t exist'])

    captured = capsys.readouterr().out

    assert captured == "No experiment found\n"


def test_exp_name(clean_db, three_experiments, monkeypatch, capsys):
    """Test that only the specified experiment is printed."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-v1\n"


def test_exp_name_with_child(clean_db, three_experiments, monkeypatch, capsys):
    """Test that only the specified experiment is printed, and with its child."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list', '--name', 'test_double_exp'])

    captured = capsys.readouterr().out

    assert captured == """\
 test_double_exp-v1┐
                   └test_double_exp_child-v1
"""


def test_exp_name_child(clean_db, three_experiments, monkeypatch, capsys):
    """Test that only the specified child experiment is printed."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list', '--name', 'test_double_exp_child'])

    captured = capsys.readouterr().out

    assert captured == " test_double_exp_child-v1\n"


def test_exp_same_name(clean_db, two_experiments_same_name, monkeypatch, capsys):
    """Test that two experiments with the same name and different versions are correctly printed."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == """\
 test_single_exp-v1┐
                   └test_single_exp-v2
"""


def test_exp_family_same_name(clean_db, three_experiments_family_same_name, monkeypatch, capsys):
    """Test that two experiments with the same name and different versions are correctly printed
    even when one of them has a child.
    """
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == """\
                   ┌test_single_exp-v2
 test_single_exp-v1┤
                   └test_single_exp_child-v1
"""


def test_exp_family_branch_same_name(clean_db, three_experiments_branch_same_name,
                                     monkeypatch, capsys):
    """Test that two experiments with the same name and different versions are correctly printed
    even when last one has a child.
    """
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == """\
 test_single_exp-v1┐
                   └test_single_exp-v2┐
                                      └test_single_exp_child-v1
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.experiment_branch_builder.py` -- Module offering an API to solve conflicts
==============================================================================================

.. module:: experiment_branch_builder
   :platform: Unix
   :synopsis: Create a list of adapters from the conflicts between an experiment and its parent.

Conflicts between two experiments arise when those have different configuration but have the same
name. Solving these conflicts require the creation of adapters to bridge from the parent experiment
and the child experiment.

 .. seealso::

    :mod:`orion.core.evc.conflicts`
    :mod:`orion.core.evc.adapters`

"""

import logging

from orion.algo.space import Dimension
import orion.core
from orion.core.evc import conflicts
from orion.core.evc.adapters import CompositeAdapter


log = logging.getLogger(__name__)


# pylint: disable=too-many-public-methods
class ExperimentBranchBuilder:
    """Build a new configuration for the experiment based on parent config.

    Parameters
    ----------
    conflicts: Conflicts
        Object reprenting a group of conflicts
    manual_resolution: bool, optional
        Starts the prompt to resolve manually the conflicts. Use system's default if not provided.
    branch_from: str, optional
        Name of the experiment to branch from.
    algorithm_change: bool, optional
        Whether to automatically solve the algorithm conflict (change of algo config).
        Defaults to True.
    code_change_type: str, optional
        How to resolve code change automatically. Must be one of 'noeffect', 'unsure' or
        'break'.  Defaults to 'break'.
    cli_change_type: str, optional
        How to resolve cli change automatically. Must be one of 'noeffect', 'unsure' or 'break'.
        Defaults to 'break'.
    config_change_type: str, optional
        How to resolve config change automatically. Must be one of 'noeffect', 'unsure' or
        'break'.  Defaults to 'break'.

    """

    def __init__(self, conflicts, manual_resolution=None, **branching_arguments):
        # TODO: handle all other arguments
        if manual_resolution is None:
            manual_resolution = orion.core.config.evc.manual_resolution

        self.manual_resolution = manual_resolution
        self.conflicts = conflicts

        for key, value in branching_arguments.items():
            if value is None and key in orion.core.config.evc:
                branching_arguments[key] = orion.core.config.evc[key]

        self.branching_arguments = branching_arguments

        self.conflicting_config.update(branching_arguments)
        self.resolve_conflicts()

    @property
    def experiment_config(self):
        """Get configuration of the parent experiment"""
        return self.conflicts.get()[0].old_config

    @property
    def conflicting_config(self):
        """Get configuration of the child experiment"""
        return self.conflicts.get()[0].new_config

    def resolve_conflicts(self, silence_errors=True):
        """Automatically resolve conflicts unless manual-resolution is True."""
        ith_conflict = 0

        while ith_conflict < len(self.conflicts.get()):

            # TODO: Meh, ugly
            conflict = self.conflicts.conflicts[ith_conflict]

            resolution = self.conflicts.try_resolve(
                conflict, silence_errors=silence_errors,
                **conflict.get_marked_arguments(self.conflicts, **self.branching_arguments))

            if resolution and (self.manual_resolution and not resolution.is_marked):
                self.conflicts.revert(resolution)

            ith_conflict += 1

    # API section
    @property
    def is_resolved(self):
        """Return True if all the current conflicts have been resolved"""
        return self.conflicts.are_resolved

    def change_experiment_name(self, name):
        """Change the child's experiment name to `name`

        Parameters
        ----------
        name: str
           New name for the child experiment. Must be different from the parent's name

        Raises
        ------
        ValueError
            If name already exists in database for current user.
        RuntimeError
            If there is no code change conflict left to resolve.

        """
        exp_name_conflicts = self.conflicts.get_remaining([conflicts.ExperimentNameConflict])
        if not exp_name_conflicts:
            raise RuntimeError('No experiment name conflict to solve')

        self.conflicts.try_resolve(exp_name_conflicts[0], name)

    def set_code_change_type(self, change_type):
        """Set code change type

        Parameters
        ----------
        change_type: string
            One of the types defined in `orion.core.evc.adapters.CodeChange.types`.

        Raises
        ------
        ValueError
            If change_type is not in `orion.core.evc.adapters.CodeChange.types`.
        RuntimeError
            If there is no code change conflict left to resolve.

        """
        code_conflicts = self.conflicts.get_remaining([conflicts.CodeConflict])
        if not code_conflicts:
            raise RuntimeError('No code conflicts to solve')

        self.conflicts.try_resolve(code_conflicts[0], change_type=change_type)

    def set_cli_change_type(self, change_type):
        """Set cli change type

        Parameters
        ----------
        change_type: string
            One of the types defined in `orion.core.evc.adapters.CommandLineChange.types`.

        Raises
        ------
        ValueError
            If change_type is not in `orion.core.evc.adapters.CommandLineChange.types`.
        RuntimeError
            If there is no cli conflict left to resolve.

        """
        cli_conflicts = self.conflicts.get_remaining([conflicts.CommandLineConflict])
        if not cli_conflicts:
            raise RuntimeError('No command line conflicts to solve')

        self.conflicts.try_resolve(cli_conflicts[0], change_type)

    def set_script_config_change_type(self, change_type):
        """Set script config change type

        Parameters
        ----------
        change_type: string
            One of the types defined in `orion.core.evc.adapters.ScriptConfigChange.types`.

        Raises
        ------
        ValueError
            If change_type is not in `orion.core.evc.adapters.ScriptConfigChange.types`.
        RuntimeError
            If there is no script config conflict left to resolve.

        """
        script_config_conflicts = self.conflicts.get_remaining([conflicts.ScriptConfigConflict])
        if not script_config_conflicts:
            raise RuntimeError('No script\'s config conflicts to solve')

        self.conflicts.try_resolve(script_config_conflicts[0], change_type)

    def set_algo(self):
        """Set algorithm resolution

        Raises
        ------
        RuntimeError
            If there is no algorithm conflict left to resolve.

        """
        algo_conflicts = self.conflicts.get_remaining([conflicts.AlgorithmConflict])
        if not algo_conflicts:
            raise RuntimeError('No algo conflict to solve')

        self.conflicts.try_resolve(algo_conflicts[0])

    def add_dimension(self, name, default_value=Dimension.NO_DEFAULT_VALUE):
        """Add dimension with given `name`

        Only dimensions with conflict type `NewDimensionConflict` or `ChangedDimensionConflict` may
        be added.

        Parameters
        ----------
        name: str
            Name of the dimension to add
        default_value: object
            Default value for the new dimension. Defaults to `Dimension.NO_DEFAULT_VALUE`.
            If conflict is ChangedDimensionConflict, default_value is ignored.

        Raises
        ------
        ValueError
            If name is not present in non-resolved conflicts or if default_value is invalid for the
            corresponding dimension.

        """
        conflict = self.conflicts.get_remaining(
            [conflicts.NewDimensionConflict, conflicts.ChangedDimensionConflict],
            dimension_name=name)[0]

        if isinstance(conflict, conflicts.NewDimensionConflict):
            self.conflicts.try_resolve(conflict, default_value=default_value)
        else:
            self.conflicts.try_resolve(conflict)

    def remove_dimension(self, name, default_value=Dimension.NO_DEFAULT_VALUE):
        """Remove dimension with given `name`

        Only dimensions with conflict type `MissingDimensionConflict` may be added.

        Parameters
        ----------
        name: str
            Name of the dimension to add
        default_value: object
            Default value for the missing dimension. Defaults to `Dimension.NO_DEFAULT_VALUE`.

        Raises
        ------
        ValueError
            If name is not present in non-resolved conflicts or if default_value is invalid for the
            corresponding dimension.

        """
        conflict = self.conflicts.get_remaining(
            [conflicts.MissingDimensionConflict], dimension_name=name)[0]

        self.conflicts.try_resolve(conflict, default_value=default_value)

    def rename_dimension(self, old_name, new_name):
        """Rename dimension `old_name` to `new_name`

        Only dimensions with conflict type `MissingDimensionConflict` may be renamed,
        and it can only be renamed to dimensions with conflict type `NewDimensionConflict`.

        Parameters
        ----------
        old_name: str
            Name of the dimension to rename
        new_name: str
            Name of the target dimension

        Raises
        ------
        ValueError
            If name is not present in non-resolved conflicts.

        Notes
        -----
        This may create a new conflict if the target dimension has a different prior.

        """
        potential_conflicts = self.conflicts.get_remaining(
            [conflicts.MissingDimensionConflict], dimension_name=old_name)

        assert len(potential_conflicts) == 1, ("Many missing dimensions with the same name: "
                                               "{}".format(", ".join(potential_conflicts)))

        old_dim_conflict = potential_conflicts[0]

        potential_conflicts = self.conflicts.get_remaining(
            [conflicts.NewDimensionConflict], dimension_name=new_name)

        assert len(potential_conflicts) == 1, ("Many new dimensions with the same name: "
                                               "{}".format(", ".join(potential_conflicts)))

        new_dim_conflict = potential_conflicts[0]

        self.conflicts.try_resolve(old_dim_conflict, new_dimension_conflict=new_dim_conflict)

    def reset(self, name):
        """Revert a resolution and reset its corresponding conflicts

        Parameters
        ----------
        name: str
            String representing the resolution as provided in the prompt

        Raises
        ------
        ValueError
            If name does not correspond to one of the current resolutions.

        Notes
        -----
        Side-effect conflicts generated by a reverted resolution will be deleted.

        """
        self.conflicts.revert(name)

    def create_adapters(self):
        """Return a list of adapters for every resolution"""
        adapters = []
        for resolution in self.conflicts.get_resolutions():
            adapters += resolution.get_adapters()

        return CompositeAdapter(*adapters)



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines,arguments-differ
"""
:mod:`orion.core.evc.conflicts` -- Description and resolution of configuration conflicts
========================================================================================

.. module:: conflicts
   :platform: Unix
   :synopsis: Detection and description of configuration conflicts and their resolutions

Conflicts between a parent experiment and a child configuration exist in many different forms. This
module provides the function `detect_conflicts` to automatically detect them. Any conflict type
which inherits from class `Conflict` is used to detect corresponding conflicts. These conflicts
can than be used to generate resolutions which will generate adapters to make trials of
both experiments compatible when applicable.

Conflicts objects know how to resolve themselves but may lack information for doing so.  For
instance `ExperimentNameConflict` knows it may only be resolved with an `ExperimentNameResolution`,
but it cannot do so unless it is given a new name to instantiate the resolution.

Conflict objects may build different resolutions based on the input given. For instance
`MissingDimensionConflict` may instantiate a `RenameDimensionResolution` if a `NewDimensionConflict`
is passed to `try_resolve`, otherwise the resolution will be `RemoveDimensionResolution`.

In short, conflict knows:

#. How to detect themselves in pair old_config, new_config (`Conflict.detect()`)
#. How to resolve themselves (but may lack information for doing so) (`conflict.try_resolve()`)
#. How to build a diff for user interface (`conflict.diff`)
#. How to build a string to represent themselves in user interface (`repr(conflict)`)
#. How to find resolutions markers and their corresponding arguments
   (`conflict.get_marked_arguments`)

while resolution knows:

#. How to create adapters (`resolution.get_adapters()`)
#. How to find marked arguments for themselves (determining if this resolution was marked by user)
   (`resolution.find_marked_argument()`)
#. How to revert themselves and resetting the corresponding conflicts (`resolution.revert()`)
#. How to validate themselves on instantiation (`resolution.validate()`)
#. How to build a string to represent themselves in user interface (`repr(resolution)`)
   (note: this string is the one a user would use to mark the resolution in command line or in
   configuration file)

The class Conflicts is provided for convenience. It provides interface to register, fetch or
deprecate (remove) conflicts. Additionaly, it provides a helper method with wraps `try_resolve` of
the Conflict objects, handling invalid resolution errors or additional new conflicts
created by resolutions. For instance, a `RenameDimensionResolution` may create a new
`ChangedDimensionConflict` if the new name is associated to a different prior than the one
of the old name.
"""

from abc import (ABCMeta, abstractmethod)
import copy
import pprint
import traceback

from orion.algo.space import Dimension
import orion.core
from orion.core.evc import adapters
from orion.core.io.orion_cmdline_parser import OrionCmdlineParser
from orion.core.io.space_builder import SpaceBuilder
from orion.core.utils.diff import colored_diff
from orion.core.utils.format_trials import standard_param_name
from orion.storage.base import get_storage


def _create_param(dimension, default_value):
    """Create a parameter dictionary based on dimension and default_value"""
    return dict(name=dimension.name, type=dimension.type, value=default_value)


def _build_extended_user_args(config):
    """Return a list of user arguments augmented with key-value pairs found in
    user's script's configuration file.
    """
    user_args = config['metadata']['user_args']

    # No need to pass config_prefix because we have access to everything
    # required in config[metadata][parser] (parsed data)
    parser = OrionCmdlineParser()
    parser.set_state_dict(config['metadata']['parser'])

    return user_args + [standard_param_name(key) + value
                        for key, value in parser.config_file_data.items()]


def _build_space(config):
    """Build an optimization space based on given configuration"""
    space_builder = SpaceBuilder()
    space_config = config['space']
    space = space_builder.build(space_config)

    return space


def detect_conflicts(old_config, new_config, branching=None):
    """Generate a Conflicts object with all conflicts found in pair (old_config, new_config)"""
    conflicts = Conflicts()
    for conflict_class in sorted(Conflict.__subclasses__(), key=lambda cls: cls.__name__):
        for conflict in conflict_class.detect(old_config, new_config, branching):
            conflicts.register(conflict)

    return conflicts


class Conflicts(object):
    """Handler of a list of conflicts

    Registers, deprecate, resolve and fetch conflicts. Revert and fetch corresponding resolutions.

    The helper method `try_resolve` wraps `Conflict.try_resolve` objects,
    handling invalid resolution errors messages and adding to its list additional new conflicts
    created by resolutions.

    Attributes
    ----------
    conflicts: list of `Conflict`
        List of conflicts which may be resolved or not.

    """

    def __init__(self):
        """Initialize empty list of conflicts"""
        self.conflicts = []

    def register(self, conflict):
        """Add a new conflict to the list of conflicts"""
        self.conflicts.append(conflict)

    def revert(self, resolution_or_name):
        """Revert a resolution and deprecate conflicts if applicable"""
        name = str(resolution_or_name)
        resolution_strings = list(map(str, (c.resolution for c in self.get_resolved())))
        resolution = self.get_resolved()[resolution_strings.index(name)].resolution
        self.deprecate(resolution.revert())

    def _get(self, callback=None):
        """Fetch conflicts for which callback return True if callback is given, else return all"""
        return [conflict for conflict in self.conflicts
                if (callback is None or callback(conflict))]

    def get(self, types=(), dimension_name=None, callback=None):
        """Fetch conflicts

        Parameters
        ----------
        types: tuple of Conflict types
            List of conflict types to fetch
        dimension_name: None or string
            name of a dimension to fetch. If not None, will raise an error if no conflict found.
        callback: None or callable object
            If not None, only conflict for which the callback return True will be returned by
            get()

        Raises
        ------
        ValueError
            If argument dimension_name is not None and no conflict is found.

        """
        def wrap(types, dimension_name, callback):
            """Wrap types, dimension_name and callback inside another callback"""
            def _callback(conflict):
                if callback is not None and not callback(conflict):
                    return False
                if types and not isinstance(conflict, tuple(types)):
                    return False
                if (dimension_name is not None and
                        (not hasattr(conflict, "dimension") or
                         standard_param_name(conflict.dimension.name) != dimension_name)):
                    return False

                return True
            return _callback

        found_conflicts = self._get(wrap(types, dimension_name, callback))

        if dimension_name is not None and not found_conflicts:
            raise ValueError("Dimension name \'{}\' not found in conflicts".format(dimension_name))

        return found_conflicts

    def get_remaining(self, types=(), dimension_name=None, callback=None):
        """Fetch non resolved conflicts

        .. note::

            See :meth:`orion.core.evc.conflict.Conflicts.get` for more information.

        """
        def _is_not_resolved(conflict):
            return not conflict.is_resolved and (callback is None or callback(conflict))
        return self.get(types, dimension_name, callback=_is_not_resolved)

    def get_resolved(self, types=(), dimension_name=None, callback=None):
        """Fetch resolved conflicts

        .. note::

            See :meth:`orion.core.evc.conflict.Conflicts.get` for more information.

        """
        def _is_resolved(conflict):
            return conflict.is_resolved and (callback is None or callback(conflict))
        return self.get(types, dimension_name, callback=_is_resolved)

    def get_resolutions(self, types=(), dimension_name=None, callback=None):
        """Fetch resolutions

        Iterate over resolved conflicts and return their resolutions

        .. note::

            Some resolutions resolve many conflicts. This method only returns unique resolutions.

        .. note::

            See :meth:`orion.core.evc.conflict.Conflicts.get` for more information.

        """
        resolutions = set()
        for conflict in self.get_resolved(types, dimension_name, callback):
            if conflict.resolution not in resolutions:
                resolutions.add(conflict.resolution)
                yield conflict.resolution

    # API section
    @property
    def are_resolved(self):
        """Return True if all the current conflicts have been resolved"""
        return all(conflict.is_resolved for conflict in self.conflicts)

    def deprecate(self, conflicts):
        """Remove given conflicts from the internal list of conflicts"""
        for conflict in conflicts:
            self.conflicts.pop(self.conflicts.index(conflict))

    def try_resolve(self, conflict, *args, **kwargs):
        """Wrap call to conflict.try_resolve

        Catch errors on `conflict.try_resolve` and print traceback if argument `silence_errors` is
        False.

        Resolutions may generate side-effect conflicts. In such case, they are added to interval's
        list of conflicts.

        Parameters
        ----------
        conflict: `orion.ore.evc.conflicts.Conflict`
            Conflict object to call `try_resolve`.
        silence_errors: bool
            If True, errors raised on execution of conflict.try_resolve will be catched and
            silenced. If False, errors will be catched and traceback will be printed before
            methods return None. Defaults to False
        *args:
            Arguments to pass to `conflict.try_resolve`
        **kwargs:
            Keyword arguments to pass to `conflict.try_resolve`

        """
        silence_errors = kwargs.pop("silence_errors", False)
        try:
            resolution = conflict.try_resolve(*args, **kwargs)
        except KeyboardInterrupt:
            raise
        except Exception:  # pylint:disable=broad-except
            conflict.resolution = None
            conflict._is_resolved = None  # pylint:disable=protected-access
            if not silence_errors:
                print(traceback.format_exc())
            return None

        if resolution:
            self.conflicts += resolution.new_conflicts

        return resolution


class Conflict(object, metaclass=ABCMeta):
    """Representation of a conflict between two configurations

    This object is used to embody a conflict during a branching event and provides means to
    resolve itself and to represent itself in user interface.

    A conflict must provide implementations of:

    #. `detect()` -- How it is detected in a pair (old_config, new_config).
    #. `try_resolve()` -- How to resolve itself.
    #. `__repr__()` -- How to represent itself in user interface.

    Additionaly, it may also provide implementations of:

    #. `diff()` -- How to compute diff string.
    #. `get_marked_arguments()` -- How to find resolutions markers and their corresponding arguments
       in `new_config`.

    Attributes
    ----------
    old_config: dict
        Configuration of the parent experiment
    new_config: dict
        Configuration of the child experiment
    resolution: None or `orion.core.evc.conflicts.Resolution`
        None if not resolved or a `Resolution` object. Note that deprecated
        conflicts may be marked as resolved with `_is_resolved = True` even though
        `resolution` is `None`.

    """

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect all conflicts in given pair (old_config, new_config) and return a list of them
        :param branching_config:
        """
        pass

    def __init__(self, old_config, new_config):
        """Initialize conflict as non-resolved"""
        self.old_config = old_config
        self.new_config = new_config
        self._is_resolved = False
        self.resolution = None

    @property
    def is_resolved(self):
        """Return True if conflict is set as resolved or if it has a resolution"""
        return self._is_resolved or self.resolution is not None

    # pylint:disable=unused-argument,no-self-use
    def get_marked_arguments(self, conflicts, **branching_kwargs):
        """Return arguments from marked resolutions in new configuration

        Some conflicts may be passed arguments with their marker to automate conflict resolution.
        For instance, a renaming resolution requires the name of a new dimension. In such case, the
        conflict for `missing_dim` would find `~missing_dim~>new_dim` in the user
        command line of configuration arguments, fetch `new_dim` from `conflicts` and return the
        dictionary of arguments `{'new_dimension_conflict': new_dimension_conflict}`.

        Parameters
        ----------
        conflicts: `orion.core.evc.conflicts.Conflicts`
            Handler of the list of conflicts.

        Returns
        -------
        dict
            Marked arguments for `conflict.try_resolve()`, which may latter be passed as
            `**kwargs`.

        """
        return {}

    @abstractmethod
    def try_resolve(self):
        """Try to create a resolution

        Conflict is then marked as resolved and its attribute `resolution` now points to the
        resolution.

        Returns
        -------
        None or `orion.core.evc.conflicts.Resolution`
            Returns None if the conflict is already resolved, otherwise
            it returns a resolution object if it is successful.

        Raises
        ------
        ValueError
            If the resolution cannot be created without arguments or if the arguments passed are not
            valid. This is specific to each child of `Conflict`

        """
        pass

    @property
    def diff(self):
        """Produce human-readable differences

        Returns
        -------
        None or str
            Returns None if the conflict cannot produce diffs, otherwise it returns
            the diff as a string (possibly multi-line).

        """
        return None

    # def get_hint(self):
    #     """Return a possible resolution as a string for user interface"""
    #     resolution = self.try_resolve()
    #     hint = "Try 'set {}'".format(str(resolution))
    #     resolution.revert()
    #     return hint

    @abstractmethod
    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        pass


class Resolution(object, metaclass=ABCMeta):
    """Representation of a resolution for a conflict between two configurations

    This object is used to embody a resolution of a conflict during a branching event and
    provides means to validate itself, produce side-effect conflicts, detect corresponding user
    markers, produce corresponding adapters and represent itself in user interface.

    The string representing a resolution is precisely what a user should type in command-line
    to resolve automatically a conflict.

    A resolution must provide implementations of:

    #. `get_adapters()` --  How to adapt trials from the two experiments.
    #. `__repr__()` -- How to represent itself in user interface. Note: this should correspond to
        what user should enter in command-line for automatic resolution.

    Additionaly, it may also provide implementations of:

    #. `revert()` -- How to revert the resolution and reset corresponding conflicts
    #. `_validate()` -- How to validate if arguments for the resolution are valid.
    #. `find_marked_argument()` -- How to find marked arguments in commandline call or script
        config

    Note that resolutions do not modify the configuration, with the exception of experiment name
    resolution, hence there is no support for diffs inside resolutions. The only diffs are between
    the two configurations, hence they are defined inside the conflicts.

    Attributes
    ----------
    conflict: `orion.core.evc.conflicts.Conflict`
        The conflict which is resolved by this resolution.
    new_conflicts: list of `orion.core.evc.conflicts.Conflict`
        The side-effect conflicts cause by this resolution.
    MARKER: None or string
        The special marker if resolution is intended for dimension conflicts, otherwise None
    ARGUMENT: None or string
        The command-line argument if the resolution is not intended for dimension conflicts.

    """

    MARKER = None
    ARGUMENT = None

    def __init__(self, conflict):
        """Initialize resolution and mark conflict as resolved"""
        self.conflict = conflict
        self.new_conflicts = []

        conflict.resolution = self

    def validate(self, *args, **kwargs):
        """Wrap validate method to revert resolution on invalid arguments"""
        try:
            self._validate(*args, **kwargs)
        except Exception:
            self.revert()
            raise

    def _validate(self, *args, **kwargs):
        """Validate arguments and raise a ValueError if they are invalid"""
        pass

    @classmethod
    def namespace(cls):
        """Return namespace corresponding to self.ARGUMENT

        ARGUMENT is a command line argument, thus something in the style of `--code-change-type`.
        When arguments are passed they are saved in namespace in the style of `code_change_type`.
        This property converts command-line style to namespace style.
        """
        if not cls.ARGUMENT:
            return None

        return cls.ARGUMENT.lstrip("-").replace("-", "_")

    def revert(self):
        """Reset conflict as well as side-effect conflicts and return the latter for deprecation"""
        self.conflict.resolution = None
        return []

    @abstractmethod
    def get_adapters(self):
        """Return adapters corresponding to the resolution"""
        pass

    @abstractmethod
    def __repr__(self):
        """Representation of the resolution as it should be provided in command line of
        configuration file by the user
        """
        pass

    def find_marked_argument(self):
        """Find commandline argument on configuration argument which marks this
        type of resolution for automatic resolution
        """
        new_config = self.conflict.new_config
        marked_argument = None
        if self.MARKER:
            for arg in _build_extended_user_args(new_config):
                if arg.lstrip("-").startswith(self.prefix):
                    marked_argument = arg
                    break
        else:
            marked_argument = new_config.get(self.namespace(), None)

        return marked_argument

    @property
    def is_marked(self):
        """If this resolution is specifically marked in commandline arguments or configuration
        arguments
        """
        return self.find_marked_argument() not in [None, False]


class NewDimensionConflict(Conflict):
    """Representation of a new dimension conflict

    Attributes
    ----------
    dimension: `orion.algo.space.Dimension`
        Dimension object which is defined in new_config but not in old_config.
    prior: string
        String representing the prior of the dimension

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect all new dimensions in `new_config` based on `old_config`
        :param branching_config:
        """
        old_space = _build_space(old_config)
        new_space = _build_space(new_config)
        for name, dim in new_space.items():
            new_prior = dim.get_prior_string()
            if name not in old_space:
                yield cls(old_config, new_config, dim, new_prior)

    def __init__(self, old_config, new_config, dimension, prior):
        """Initialize conflict as non-resolved"""
        super(NewDimensionConflict, self).__init__(old_config, new_config)
        self.dimension = dimension
        self.prior = prior

    def try_resolve(self, default_value=Dimension.NO_DEFAULT_VALUE):
        """Try to create a resolution AddDimensionResolution

        Parameters
        ----------
        default_value: object
            Default value for the new dimension. Defaults to `Dimension.NO_DEFAULT_VALUE`.

        Raises
        ------
        ValueError
            If default_value is invalid for the corresponding dimension.

        """
        if self.is_resolved:
            return None

        return self.AddDimensionResolution(self, default_value)

    @property
    def diff(self):
        """Produce human-readable differences"""
        return colored_diff("", self.dimension.get_string())

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        return "New {0}".format(standard_param_name(self.dimension.name))

    class AddDimensionResolution(Resolution):
        """Representation of a new dimension resolution

        Attributes
        ----------
        default_value: object
            Default value for the new dimension.

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        MARKER = "~+"

        def __init__(self, conflict, default_value=Dimension.NO_DEFAULT_VALUE):
            """Initialize resolution and mark conflict as resolved

            Parameters
            ----------
            conflict: `orion.core.evc.conflicts.Conflict`
                The conflict which is resolved by this resolution.
            default_value: object
                Default value for the new dimension. Defaults to `Dimension.NO_DEFAULT_VALUE`.
                If `Dimension.NO_DEFAULT_VALUE`, default_value from corresponding dimension will be
                used.

            Raises
            ------
            ValueError
                If default_value is invalid for the corresponding dimension.

            """
            super(NewDimensionConflict.AddDimensionResolution, self).__init__(conflict)
            if default_value is Dimension.NO_DEFAULT_VALUE:
                default_value = conflict.dimension.default_value
            else:
                default_value = conflict.dimension.cast(default_value)

            self.validate(default_value)

            self.default_value = default_value

        def _validate(self, default_value):
            """Validate default value is NO_DEFAULT_VALUE or is in dimension's interval"""
            if ((default_value is not Dimension.NO_DEFAULT_VALUE) and
                    (default_value not in self.conflict.dimension)):
                raise ValueError(
                    "Default value `{}` is outside of dimension's prior interval `{}`".format(
                        default_value, self.conflict.prior))

        def get_adapters(self):
            """Return DimensionAddition adapter"""
            default_param = _create_param(self.conflict.dimension, self.default_value)
            return [adapters.DimensionAddition(default_param)]

        @property
        def prefix(self):
            """Build the prefix including the marker"""
            return '{0}{1}'.format(standard_param_name(self.conflict.dimension.name),
                                   self.MARKER)

        @property
        def new_prior(self):
            """Build the new prior string, including the default value"""
            tmp_dim = copy.deepcopy(self.conflict.dimension)
            tmp_dim._default_value = self.default_value  # pylint:disable=protected-access
            return tmp_dim.get_prior_string()

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return '{0}{1}'.format(self.prefix, self.new_prior)


class ChangedDimensionConflict(Conflict):
    """Representation of a changed prior conflict

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect all changed dimensions in `new_config` based on `old_config`
        :param branching_config:
        """
        old_space = _build_space(old_config)
        new_space = _build_space(new_config)
        for name, dim in new_space.items():
            if name not in old_space:
                continue

            new_prior = dim.get_prior_string()
            old_prior = old_space[name].get_prior_string()

            if new_prior != old_prior:
                yield cls(old_config, new_config, dim, old_prior, new_prior)

    def __init__(self, old_config, new_config, dimension, old_prior, new_prior):
        """Initialize conflict as non-resolved"""
        super(ChangedDimensionConflict, self).__init__(old_config, new_config)
        self.dimension = dimension
        self.old_prior = old_prior
        self.new_prior = new_prior

    def try_resolve(self):
        """Try to create a resolution ChangeDimensionResolution"""
        if self.is_resolved:
            return None

        return self.ChangeDimensionResolution(self)

    @property
    def diff(self):
        """Produce human-readable differences"""
        return colored_diff(self.old_prior, self.new_prior)

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        return "{0}~{1} != {0}~{2}".format(standard_param_name(self.dimension.name),
                                           self.old_prior, self.new_prior)

    class ChangeDimensionResolution(Resolution):
        """Representation of a changed prior resolution

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        MARKER = "~+"

        def get_adapters(self):
            """Return DimensionPriorChange adapter"""
            return [adapters.DimensionPriorChange(
                self.conflict.dimension.name, self.conflict.old_prior, self.conflict.new_prior)]

        @property
        def prefix(self):
            """Build the new prior string, including the default value"""
            return '{0}{1}'.format(standard_param_name(self.conflict.dimension.name),
                                   self.MARKER)

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return "{0}{1}".format(self.prefix, self.conflict.new_prior)


class MissingDimensionConflict(Conflict):
    """Representation of a new dimension conflict

    Attributes
    ----------
    dimension: `orion.algo.space.Dimension`
        Dimension object which is defined in new_config but not in old_config.
    prior: string
        String representing the prior of the dimension

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect all missing dimensions in `new_config` based on `old_config`
        :param branching_config:
        """
        for conflict in NewDimensionConflict.detect(new_config, old_config):
            yield cls(old_config, new_config, conflict.dimension, conflict.prior)

    def __init__(self, old_config, new_config, dimension, prior):
        """Initialize conflict as non-resolved"""
        super(MissingDimensionConflict, self).__init__(old_config, new_config)
        self.dimension = dimension
        self.prior = prior

    def get_marked_arguments(self, conflicts, **branching_kwargs):
        """Find and return marked arguments for remove or rename resolution

        .. seealso::

            :meth:`orion.core.evc.conflicts.Conflict.get_marked_arguments`

        """
        marked_remove_arguments = self.get_marked_remove_arguments(conflicts)

        if marked_remove_arguments:
            return marked_remove_arguments

        return self.get_marked_rename_arguments(conflicts)

    # pylint:disable=unused-argument
    def get_marked_remove_arguments(self, conflicts):
        """Find and return marked arguments for remove resolution

        .. seealso::

            :meth:`orion.core.evc.conflicts.Conflict.get_marked_arguments`

        """
        if self.is_resolved:
            return {}

        remove_dimension_resolution = copy.deepcopy(self).try_resolve()

        if not remove_dimension_resolution:
            return {}

        arguments = remove_dimension_resolution.find_marked_argument()
        if arguments:
            new_default_value = arguments.split(
                MissingDimensionConflict.RemoveDimensionResolution.MARKER)[1]

            if not new_default_value:
                new_default_value = Dimension.NO_DEFAULT_VALUE
            else:
                new_default_value = self.dimension.cast(new_default_value)

            return {'default_value': new_default_value}
        return {}

    def get_marked_rename_arguments(self, conflicts):
        """Find and return marked arguments for rename resolution

        .. seealso::

            :meth:`orion.core.evc.conflicts.Conflict.get_marked_arguments`

        """
        new_dimension_conflicts = conflicts.get([NewDimensionConflict])
        if not new_dimension_conflicts:
            return {}

        resolution = copy.deepcopy(self).try_resolve(
            new_dimension_conflict=copy.deepcopy(new_dimension_conflicts[0]))

        if not resolution:
            return {}

        arguments = resolution.find_marked_argument()

        if arguments:
            new_dimension_name = "~>".join(arguments.split('~>')[1:])

            try:
                conflict = conflicts.get([NewDimensionConflict],
                                         dimension_name=new_dimension_name)[0]
            except ValueError as e:
                if "Dimension name '{}' not found".format(new_dimension_name) not in str(e):
                    return {}

                raise

            if conflict.is_resolved:
                conflicts.revert(str(conflict.resolution))

            return {'new_dimension_conflict': conflict}

        return {}

    def try_resolve(self, new_dimension_conflict=None, default_value=Dimension.NO_DEFAULT_VALUE):
        """Try to create a resolution RenameDimensionResolution of RemoveDimensionResolution

        Parameters
        ----------
        new_dimension_conflict: None or `orion.core.evc.conflicts.NewDimensionConflict`
            Dimension used for a rename resolution. If None, a remove resolution will be created
            instead.
        default_value: object
            Default value for the missing dimension. Defaults to `Dimension.NO_DEFAULT_VALUE`.
            If `Dimension.NO_DEFAULT_VALUE`, default_value from corresponding dimension will be
            used. This argument is ignored if new_dimension_conflict is not None.

        Raises
        ------
        ValueError
            If default_value is invalid for the corresponding dimension.

        """
        if self.is_resolved:
            return None

        if new_dimension_conflict:
            return MissingDimensionConflict.RenameDimensionResolution(self, new_dimension_conflict)

        return MissingDimensionConflict.RemoveDimensionResolution(self, default_value)

    @property
    def diff(self):
        """Produce human-readable differences"""
        return colored_diff(self.dimension.get_string(), "")

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        return "Missing {0}".format(standard_param_name(self.dimension.name))

    class RenameDimensionResolution(Resolution):
        """Representation of a rename dimension resolution

        Attributes
        ----------
        new_dimension_conflict: `orion.core.evc.conflicts.NewDimensionConflict`
            New dimension to rename to.

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        MARKER = "~>"

        def __init__(self, conflict, new_dimension_conflict):
            """Initialize resolution and mark conflict as resolved

            .. note::

                Will create a side-effect conflict if the new dimension have a different prior than
                the old dimension.

            Parameters
            ----------
            new_dimension_conflict: `orion.core.evc.conflicts.NewDimensionConflict`
                Dimension used for a rename resolution.

            """
            super(MissingDimensionConflict.RenameDimensionResolution, self).__init__(conflict)

            self.new_dimension_conflict = new_dimension_conflict
            new_dimension_conflict.resolution = self

            if self.conflict.prior != new_dimension_conflict.prior:
                changed_dimension_conflict = ChangedDimensionConflict(
                    self.conflict.old_config,
                    self.conflict.new_config,
                    new_dimension_conflict.dimension, self.conflict.prior,
                    new_dimension_conflict.prior)

                self.new_conflicts.append(changed_dimension_conflict)

        def revert(self):
            """Reset conflict as well as side-effect conflicts and return the latter for
            deprecation
            """
            self.conflict.resolution = None
            self.new_dimension_conflict.resolution = None

            deprecated_conflicts = self.new_conflicts
            if deprecated_conflicts:
                deprecated_conflicts[0]._is_resolved = True  # pylint:disable=protected-access

            self.new_conflicts = []

            return deprecated_conflicts

        def get_adapters(self):
            """Return DimensionRenaming adapter"""
            return [adapters.DimensionRenaming(
                self.conflict.dimension.name, self.new_dimension_conflict.dimension.name)]

        @property
        def prefix(self):
            """Build the new prior string, including the default value"""
            return "{0}{1}".format(
                standard_param_name(self.conflict.dimension.name),
                self.MARKER)

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return "{0}{1}".format(
                self.prefix, standard_param_name(self.new_dimension_conflict.dimension.name))

    class RemoveDimensionResolution(Resolution):
        """Representation of a remove dimension resolution

        Attributes
        ----------
        default_value: object
            Default value for the missing dimension.

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        MARKER = "~-"

        def __init__(self, conflict, default_value=Dimension.NO_DEFAULT_VALUE):
            """Initialize resolution and mark conflict as resolved

            Parameters
            ----------
            conflict: `orion.core.evc.conflicts.Conflict`
                The conflict which is resolved by this resolution.
            default_value: object
                Default value for the missing dimension. Defaults to `Dimension.NO_DEFAULT_VALUE`.
                If `Dimension.NO_DEFAULT_VALUE`, default_value from corresponding dimension will be
                used.

            Raises
            ------
            ValueError
                If default_value is invalid for the corresponding dimension.

            """
            super(MissingDimensionConflict.RemoveDimensionResolution, self).__init__(conflict)
            if default_value is Dimension.NO_DEFAULT_VALUE:
                default_value = conflict.dimension.default_value
            else:
                default_value = self.conflict.dimension.cast(default_value)

            self.validate(default_value)
            self.default_value = default_value

        def _validate(self, default_value):
            """Validate default value is NO_DEFAULT_VALUE or is in dimension's interval"""
            if ((default_value is not Dimension.NO_DEFAULT_VALUE) and
                    (default_value not in self.conflict.dimension)):
                raise ValueError(
                    "Default value `{}` is outside of dimension's prior interval `{}`".format(
                        default_value, self.conflict.prior))

        def get_adapters(self):
            """Return DimensionDeletion adapter"""
            param = _create_param(self.conflict.dimension, self.default_value)
            return [adapters.DimensionDeletion(param)]

        @property
        def prefix(self):
            """Build the new prior string, including the default value"""
            return '{0}{1}'.format(standard_param_name(self.conflict.dimension.name),
                                   self.MARKER)

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            string = self.prefix
            if self.default_value is not Dimension.NO_DEFAULT_VALUE:
                string += "{}".format(repr(self.default_value))
            return string


class AlgorithmConflict(Conflict):
    """Representation of an algorithm configuration conflict

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect if algorithm definition in `new_config` differs from `old_config`
        :param branching_config:
        """
        if old_config['algorithms'] != new_config['algorithms']:
            yield cls(old_config, new_config)

    def try_resolve(self):
        """Try to create a resolution AlgorithmResolution"""
        if self.is_resolved:
            return None

        return self.AlgorithmResolution(self)

    @property
    def diff(self):
        """Produce human-readable differences"""
        return colored_diff(
            pprint.pformat(self.old_config['algorithms']),
            pprint.pformat(self.new_config['algorithms']))

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        # TODO: select different subset rather than printing the old dict
        return "{0}\n   !=\n{1}".format(
            pprint.pformat(self.old_config['algorithms']),
            pprint.pformat(self.new_config['algorithms']))

    class AlgorithmResolution(Resolution):
        """Representation of an algorithn configuration resolution

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        ARGUMENT = "--algorithm-change"

        def get_adapters(self):
            """Return AlgorithmChange adapter"""
            return [adapters.AlgorithmChange()]

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return "{0}".format(self.ARGUMENT)


class CodeConflict(Conflict):
    """Representation of code change conflict

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect if commit hash in `new_config` differs from `old_config`
        :param branching_config:
        """
        old_hash_commit = old_config['metadata'].get('VCS', None)
        new_hash_commit = new_config['metadata'].get('VCS')

        ignore_code_changes = branching_config is not None and \
            branching_config.get('ignore_code_changes', False)
        if not ignore_code_changes and new_hash_commit and old_hash_commit != new_hash_commit:
            yield cls(old_config, new_config)

    def get_marked_arguments(self, conflicts, code_change_type=None, **branching_kwargs):
        """Find and return marked arguments for code change conflict

        .. seealso::

            :meth:`orion.core.evc.conflicts.Conflict.get_marked_arguments`

        """
        change_type = self.new_config.get(self.CodeResolution.namespace())

        if change_type:
            return dict(change_type=change_type)

        if code_change_type is None:
            code_change_type = orion.core.config.evc.code_change_type

        return dict(change_type=code_change_type)

    def try_resolve(self, change_type=None):
        """Try to create a resolution CodeResolution

        Parameters
        ----------
        change_type: None or string
            One of the types defined in `orion.core.evc.adapters.CodeChange.types`.

        Raises
        ------
        ValueError
            If change_type is not in `orion.core.evc.adapters.CodeChange.types`.

        """
        if self.is_resolved:
            return None

        return self.CodeResolution(self, change_type)

    @property
    def diff(self):
        """Produce human-readable differences"""
        return colored_diff(
            self.old_config['metadata'].get('VCS', None),
            self.new_config['metadata'].get('VCS'))

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        return "Old hash commit '{0}'  != new hash commit '{1}'".format(
            pprint.pformat(self.old_config['metadata'].get('VCS', None)).replace("\n", ""),
            pprint.pformat(self.new_config['metadata'].get('VCS')).replace("\n", ""))

    class CodeResolution(Resolution):
        """Representation of an code change resolution

        Attributes
        ----------
        conflict: `orion.core.evc.conflicts.Conflict`
            The conflict which is resolved by this resolution.
        change_type: string
            One of the types defined in `orion.core.evc.adapters.CodeChange.types`.

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        ARGUMENT = "--code-change-type"

        def __init__(self, conflict, change_type):
            """Initialize resolution and mark conflict as resolved

            Parameters
            ----------
            conflict: `orion.core.evc.conflicts.Conflict`
                The conflict which is resolved by this resolution.
            change_type: string
                One of the types defined in `orion.core.evc.adapters.CodeChange.types`.

            Raises
            ------
            ValueError
                If change_type is not in `orion.core.evc.adapters.CodeChange.types`.

            """
            super(CodeConflict.CodeResolution, self).__init__(conflict)

            self.validate(change_type)
            self.type = change_type

        def _validate(self, change_type):
            """Validate change_type is in `orion.core.evc.adapters.CodeChange.types`"""
            adapters.CodeChange.validate(change_type)

        def get_adapters(self):
            """Return CodeChange adapter"""
            return [adapters.CodeChange(self.type)]

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return "{0} {1}".format(self.ARGUMENT, self.type)


class CommandLineConflict(Conflict):
    """Representation of commandline change conflict

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    # pylint: disable=unused-argument
    @classmethod
    def get_nameless_args(cls, config, user_script_config=None,
                          non_monitored_arguments=None, **kwargs):
        """Get user's commandline arguments which are not dimension definitions"""
        # Used python API
        if 'parser' not in config['metadata']:
            return ""

        if user_script_config is None:
            user_script_config = orion.core.config.worker.user_script_config
        if non_monitored_arguments is None:
            non_monitored_arguments = orion.core.config.evc.non_monitored_arguments

        parser = OrionCmdlineParser(user_script_config)
        parser.set_state_dict(config['metadata']['parser'])
        priors = parser.priors_to_normal()
        nameless_keys = set(parser.parser.arguments.keys()) - set(priors.keys())

        nameless_args = {key: arg for key, arg in parser.parser.arguments.items()
                         if key in nameless_keys and key not in non_monitored_arguments}

        return " ".join(" ".join([key, str(arg)]) for key, arg in
                        sorted(nameless_args.items(), key=lambda a: a[0]))

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect if command line call in `new_config` differs from `old_config`
        :param branching_config:
        """
        if branching_config is None:
            branching_config = {}
        old_nameless_args = cls.get_nameless_args(old_config, **branching_config)
        new_nameless_args = cls.get_nameless_args(new_config, **branching_config)

        if old_nameless_args != new_nameless_args:
            yield cls(old_config, new_config)

    def get_marked_arguments(self, conflicts, cli_change_type=None, **branching_kwargs):
        """Find and return marked arguments for cli change conflict

        .. seealso::

            :meth:`orion.core.evc.conflicts.Conflict.get_marked_arguments`

        """
        change_type = self.new_config.get(self.CommandLineResolution.namespace())

        if change_type:
            return dict(change_type=change_type)

        if cli_change_type is None:
            cli_change_type = orion.core.config.evc.cli_change_type

        return dict(change_type=cli_change_type)

    def try_resolve(self, change_type=None):
        """Try to create a resolution CommandLineResolution

        Parameters
        ----------
        change_type: None or string
            One of the types defined in `orion.core.evc.adapters.CommandLineChange.types`.

        Raises
        ------
        ValueError
            If change_type is not in `orion.core.evc.adapters.CommandLineChange.types`.

        """
        if self.is_resolved:
            return None

        return self.CommandLineResolution(self, change_type)

    @property
    def diff(self):
        """Produce human-readable differences"""
        return colored_diff(self.get_nameless_args(self.old_config),
                            self.get_nameless_args(self.new_config))

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        return "Old arguments \'{0}\' != new arguments \'{1}\'".format(
            self.get_nameless_args(self.old_config),
            self.get_nameless_args(self.new_config))

    class CommandLineResolution(Resolution):
        """Representation of an commandline change resolution

        Attributes
        ----------
        conflict: `orion.core.evc.conflicts.Conflict`
            The conflict which is resolved by this resolution.
        change_type: string
            One of the types defined in `orion.core.evc.adapters.CommandLineChange.types`.

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        ARGUMENT = "--cli-change-type"

        def __init__(self, conflict, change_type):
            """Initialize resolution and mark conflict as resolved

            Parameters
            ----------
            conflict: `orion.core.evc.conflicts.Conflict`
                The conflict which is resolved by this resolution.
            change_type: string
                One of the types defined in `orion.core.evc.adapters.CommandLineChange.types`.

            Raises
            ------
            ValueError
                If change_type is not in `orion.core.evc.adapters.CommandLineChange.types`.

            """
            super(CommandLineConflict.CommandLineResolution, self).__init__(conflict)

            self.validate(change_type)
            self.type = change_type

        def _validate(self, change_type):
            """Validate change_type is in `orion.core.evc.adapters.CommandLineChange.types`"""
            adapters.CommandLineChange.validate(change_type)

        def get_adapters(self):
            """Return CommandLineChange adapter"""
            return [adapters.CommandLineChange(self.type)]

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return "{0} {1}".format(self.ARGUMENT, self.type)


class ScriptConfigConflict(Conflict):
    """Representation of script configuration change conflict

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    # pylint:disable=unused-argument
    @classmethod
    def get_nameless_config(cls, config, user_script_config=None, **branching_kwargs):
        """Get configuration dict of user's script without dimension definitions"""
        # Used python API
        if 'parser' not in config['metadata']:
            return ""

        if user_script_config is None:
            user_script_config = orion.core.config.worker.user_script_config

        parser = OrionCmdlineParser(user_script_config)
        parser.set_state_dict(config['metadata']['parser'])

        nameless_config = dict((key, value)
                               for (key, value) in parser.config_file_data.items()
                               if not (isinstance(value, str) and value.startswith('orion~')))

        return nameless_config

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Detect if user's script's config file in `new_config` differs from `old_config`
        :param branching_config:
        """
        if branching_config is None:
            branching_config = {}

        old_script_config = cls.get_nameless_config(old_config, **branching_config)
        new_script_config = cls.get_nameless_config(new_config, **branching_config)

        if old_script_config != new_script_config:
            yield cls(old_config, new_config)

    def get_marked_arguments(self, conflicts, config_change_type=None, **branching_kwargs):
        """Find and return marked arguments for user's script's config change conflict

        .. seealso::

            :meth:`orion.core.evc.conflicts.Conflict.get_marked_arguments`

        """
        change_type = self.new_config.get(self.ScriptConfigResolution.namespace())

        if change_type:
            return dict(change_type=change_type)

        if config_change_type is None:
            config_change_type = orion.core.config.evc.config_change_type

        return dict(change_type=config_change_type)

    def try_resolve(self, change_type=None):
        """Try to create a resolution ScriptConfigResolution

        Parameters
        ----------
        change_type: None or string
            One of the types defined in `orion.core.evc.adapters.ScriptConfigChange.types`.

        Raises
        ------
        ValueError
            If change_type is not in `orion.core.evc.adapters.ScriptConfigChange.types`.

        """
        if self.is_resolved:
            return None

        return self.ScriptConfigResolution(self, change_type)

    @property
    def diff(self):
        """Produce human-readable differences"""
        return colored_diff(
            pprint.pformat(self.get_nameless_config(self.old_config)),
            pprint.pformat(self.get_nameless_config(self.new_config)))

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        return "Script's configuration file changed"

    class ScriptConfigResolution(Resolution):
        """Representation of a script configuration change resolution

        Attributes
        ----------
        conflict: `orion.core.evc.conflicts.Conflict`
            The conflict which is resolved by this resolution.
        change_type: string
            One of the types defined in `orion.core.evc.adapters.ScriptConfighange.types`.

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        """

        ARGUMENT = "--config-change-type"

        def __init__(self, conflict, change_type):
            """Initialize resolution and mark conflict as resolved

            Parameters
            ----------
            conflict: `orion.core.evc.conflicts.Conflict`
                The conflict which is resolved by this resolution.
            change_type: string
                One of the types defined in `orion.core.evc.adapters.ScriptConfigChange.types`.

            Raises
            ------
            ValueError
                If change_type is not in `orion.core.evc.adapters.ScriptConfigChange.types`.

            """
            super(ScriptConfigConflict.ScriptConfigResolution, self).__init__(conflict)

            self.validate(change_type)
            self.type = change_type

        def _validate(self, change_type):
            """Validate change_type is in `orion.core.evc.adapters.ScriptConfigChange.types`"""
            adapters.ScriptConfigChange.validate(change_type)

        def get_adapters(self):
            """Return ScriptdConfigChange adapter"""
            return [adapters.ScriptConfigChange(self.type)]

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return "{0} {1}".format(self.ARGUMENT, self.type)


class ExperimentNameConflict(Conflict):
    """Representation of experiment name conflict

    .. seealso ::

        :class:`orion.core.evc.conflicts.Conflict`

    """

    @classmethod
    def detect(cls, old_config, new_config, branching_config=None):
        """Return experiment name conflict no matter what

        Branching event cannot be triggered experiment name is not the same.
        :param branching_config:
        """
        yield cls(old_config, new_config)

    def get_marked_arguments(self, conflicts, **branching_kwargs):
        """Find and return marked arguments for experiment name conflict

        .. seealso::

            :meth:`orion.core.evc.conflicts.Conflict.get_marked_arguments`

        """
        new_name = self.new_config.get(self.ExperimentNameResolution.namespace())

        if new_name:
            return dict(new_name=new_name)

        return {}

    @property
    def version(self):
        """Retrieve version of configuration"""
        return self.old_config['version']

    def try_resolve(self, new_name=None):
        """Try to create a resolution ExperimentNameResolution

        Parameters
        ----------
        new_name: None or string
            A new name for the branching experiment. A ValueError is raised if name is already in
            database.

        Raises
        ------
        ValueError
            If name already exists in database for current version.

        """
        if self.is_resolved:
            return None

        return self.ExperimentNameResolution(self, new_name)

    @property
    def diff(self):
        """Produce *no* diff"""
        return None

    def __repr__(self):
        """Reprensentation of the conflict for user interface"""
        return "Experiment name \'{0}\' already exist with version \'{1}\'".format(
            self.old_config['name'], self.version)

    class ExperimentNameResolution(Resolution):
        """Representation of an experiment name resolution

        .. seealso ::

            :class:`orion.core.evc.conflicts.Resolution`

        Attributes
        ----------
        conflict: `orion.core.evc.conflicts.Conflict`
            The conflict which is resolved by this resolution.
        new_name: string
            A new name for the branching experiment.

        """

        ARGUMENT = "--branch-to"

        def __init__(self, conflict, new_name):
            """Initialize resolution and mark conflict as resolved

            Parameters
            ----------
            conflict: `orion.core.evc.conflicts.Conflict`
                The conflict which is resolved by this resolution.
            new_name: string
                A new name for the branching experiment. A ValueError is raised if name is already
                in database with a direct child.

            Raises
            ------
            ValueError
                If name already exists in database with a direct child for current version.

            """
            super(ExperimentNameConflict.ExperimentNameResolution, self).__init__(conflict)

            self.new_name = new_name
            self.old_name = self.conflict.old_config['name']
            self.old_version = self.conflict.old_config.get('version', 1)
            self.new_version = self.old_version
            self.validate()
            self.conflict.new_config['name'] = self.new_name
            self.conflict.new_config['version'] = self.new_version

        def _validate(self):
            """Validate new_name is not in database with a direct child for current version"""
            # TODO: WARNING!!! _name_is_unique could lead to race conditions,
            # The resolution may become invalid before the branching experiment is
            # registered. What should we do in such case?
            if self.new_name is not None and self.new_name != self.old_name:
                # If we are trying to actually branch from experiment
                if not self._name_is_unique():
                    raise ValueError(
                        "Cannot branch from {} with name {} since it already exists.".format(
                            self.old_name, self.new_name))
                # Since the name changes, we reset the version count.
                self.new_version = 1

            # If the new name is the same as the old name, we are trying to increment
            # the version of the experiment.
            elif self._check_for_greater_versions():
                raise ValueError(
                    "Experiment name \'{0}\' already exist for version \'{1}\' and has children. "
                    "Version cannot be auto-incremented and a new name is required for branching."
                    .format(self.new_name, self.conflict.version))
            else:
                self.new_name = self.old_name
                self.new_version = self.conflict.old_config.get('version', 1) + 1

        def _name_is_unique(self):
            """Return True if given name is not in database for current version"""
            query = {'name': self.new_name, 'version': self.conflict.version}

            named_experiments = len(get_storage().fetch_experiments(query))
            return named_experiments == 0

        def _check_for_greater_versions(self):
            """Check if experiment has children"""
            # If we made it this far, new_name is actually the name of the parent.
            parent = self.conflict.old_config

            query = {'name': parent['name'], 'refers.parent_id': parent['_id']}
            children = len(get_storage().fetch_experiments(query))

            return bool(children)

        def revert(self):
            """Reset conflict set experiment name back to old one in new configuration"""
            self.conflict.new_config['name'] = self.old_name
            self.conflict.new_config['version'] = self.old_version
            return super(ExperimentNameConflict.ExperimentNameResolution, self).revert()

        def get_adapters(self):
            """Return no adapters, trials need to adaptation to new experiment name"""
            return []

        def __repr__(self):
            """Representation of the resolution as it should be provided in command line of
            configuration file by the user
            """
            return "{0} {1}".format(self.ARGUMENT, self.new_name)

        @property
        def is_marked(self):
            """Return True every time since the `--branch-from` argument is not used when
            incrementing version of an experiment.
            """
            return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.client.experiment`."""
import atexit
import copy
import datetime
import logging

import pytest

import orion.core
from orion.core.io.database import DuplicateKeyError
from orion.core.utils.exceptions import BrokenExperiment, SampleTimeout
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage
from orion.testing import create_experiment

config = dict(
    name='supernaekei',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'tsirif',
              'orion_version': 'XYZ',
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
    refers=dict(
        root_id='supernaekei',
        parent_id=None,
        adapter=[])
)


base_trial = {
    'experiment': 0,
    'status': 'new',  # new, reserved, suspended, completed, broken
    'worker': None,
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [],
    'params': []
}


def compare_trials(trials_a, trials_b):
    """Compare two trials by using their configuration"""
    def to_dict(trial):
        return trial.to_dict()

    assert list(map(to_dict, trials_a)) == list(map(to_dict, trials_b))


def compare_without_heartbeat(trial_a, trial_b):
    """Compare trials configuration ommiting heartbeat"""
    trial_a_dict = trial_a.to_dict()
    trial_b_dict = trial_b.to_dict()
    trial_a_dict.pop('heartbeat')
    trial_b_dict.pop('heartbeat')
    assert trial_a_dict == trial_b_dict


def test_plot_is_defined():
    """Tests plot() method is defined"""
    with create_experiment(config, base_trial) as (_, _, client):
        assert client.plot()


def test_experiment_fetch_trials():
    """Test compliance of client and experiment `fetch_trials()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        assert len(experiment.fetch_trials()) == 5
        compare_trials(experiment.fetch_trials(), client.fetch_trials())


def test_experiment_get_trial():
    """Test compliance of client and experiment `get_trial()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        assert experiment.get_trial(uid=0) == client.get_trial(uid=0)


def test_experiment_fetch_trials_by_status():
    """Test compliance of client and experiment `fetch_trials_by_status()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        compare_trials(experiment.fetch_trials_by_status('completed'),
                       client.fetch_trials_by_status('completed'))


def test_experiment_fetch_non_completed_trials():
    """Test compliance of client and experiment `fetch_noncompleted_trials()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        compare_trials(experiment.fetch_noncompleted_trials(), client.fetch_noncompleted_trials())


class TestInsert:
    """Tests for ExperimentClient.insert"""

    def test_insert_params_wo_results(self):
        """Test insertion without results without reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.insert(dict(x=100))
            assert trial.status == 'interrupted'
            assert trial.params['x'] == 100
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))

            assert client._pacemakers == {}

    def test_insert_params_with_results(self):
        """Test insertion with results without reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            timestamp = datetime.datetime.utcnow()
            trial = client.insert(dict(x=100),
                                  [dict(name='objective', type='objective', value=101)])
            assert trial.status == 'completed'
            assert trial.params['x'] == 100
            assert trial.objective.value == 101
            assert trial.end_time >= timestamp
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))
            assert client.get_trial(uid=trial.id).objective.value == 101

            assert client._pacemakers == {}

    def test_insert_params_with_results_and_reserve(self):
        """Test insertion with results and reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(
                    dict(x=100),
                    [dict(name='objective', type='objective', value=101)],
                    reserve=True)

            assert 'Cannot observe a trial and reserve it' in str(exc.value)

    def test_insert_existing_params(self):
        """Test that duplicated trials cannot be saved in storage"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(DuplicateKeyError) as exc:
                client.insert(dict(x=1))

            assert ('A trial with params {\'x\': 1} already exist for experiment supernaekei-v1' ==
                    str(exc.value))

            assert client._pacemakers == {}

    def test_insert_partial_params(self):
        """Test that trial with missing dimension that has a default value can be saved"""
        config_with_default = copy.deepcopy(config)
        config_with_default['space']['y'] = 'uniform(0, 10, default_value=5)'
        trial_with_default = copy.deepcopy(base_trial)
        trial_with_default['params'].append({'name': 'y', 'type': 'real', 'value': 1})
        with create_experiment(config_with_default, trial_with_default) as (_, experiment, client):
            trial = client.insert(dict(x=100))

            assert trial.status == 'interrupted'
            assert trial.params['x'] == 100
            assert trial.params['y'] == 5
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))

            assert client._pacemakers == {}

    def test_insert_partial_params_missing(self):
        """Test that trial with missing dimension cannot be saved"""
        config_with_default = copy.deepcopy(config)
        config_with_default['space']['y'] = 'uniform(0, 10)'
        trial_with_default = copy.deepcopy(base_trial)
        trial_with_default['params'].append({'name': 'y', 'type': 'real', 'value': 1})
        with create_experiment(config_with_default, trial_with_default) as (_, _, client):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x=1))

            assert ('Dimension y not specified and does not have a default value.' ==
                    str(exc.value))

    def test_insert_params_and_reserve(self):
        """Test that new trial is reserved properly with `reserve=True`"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.insert(dict(x=100), reserve=True)
            assert trial.status == 'reserved'
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_insert_params_fails_not_reserved(self):
        """Test that failed insertion because of duplicated trials will not reserve the original
        trial
        """
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(DuplicateKeyError):
                client.insert(dict(x=1), reserve=True)

            assert client._pacemakers == {}

    def test_insert_bad_params(self):
        """Test that bad params cannot be registered in storage"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x='bad bad bad'))

            assert ('Dimension x value bad bad bad is outside of prior uniform(0, 200)' ==
                    str(exc.value))
            assert client._pacemakers == {}

    def test_insert_params_bad_results(self):
        """Test that results with from format cannot be saved (trial is registered anyhow)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x=100), [dict(name='objective', type='bad bad bad', value=0)])

            assert 'Given type, bad bad bad, not one of: ' in str(exc.value)
            assert client._pacemakers == {}


class TestReserve:
    """Tests for ExperimentClient.reserve"""

    def test_reserve(self):
        """Test reservation of registered trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            assert trial.status != 'reserved'
            client.reserve(trial)
            assert trial.status == 'reserved'
            assert experiment.get_trial(trial).status == 'reserved'
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_reserve_dont_exist(self):
        """Verify that unregistered trials cannot be reserved."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment='idontexist', params=cfg.trials[0]['params'])
            with pytest.raises(ValueError) as exc:
                client.reserve(trial)

            assert 'Trial {} does not exist in database.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}

    def test_reserve_reserved_locally(self, caplog):
        """Verify that a trial cannot be reserved twice locally (warning, no exception)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            assert trial.status != 'reserved'
            client.reserve(trial)
            with caplog.at_level(logging.WARNING):
                client.reserve(trial)

            assert 'Trial {} is already reserved.'.format(trial.id) == caplog.records[-1].message

            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_reserve_reserved_remotely(self):
        """Verify that a trial cannot be reserved if already reserved by another process"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            assert trial.status == 'interrupted'
            client.reserve(trial)
            remote_pacemaker = client._pacemakers.pop(trial.id)
            assert experiment.get_trial(trial).status == 'reserved'

            trial = Trial(**cfg.trials[1])
            assert trial.status == 'interrupted'
            with pytest.raises(RuntimeError) as exc:
                client.reserve(trial)

            assert 'Could not reserve trial {}.'.format(trial.id) == str(exc.value)

            assert trial.status == 'interrupted'
            assert experiment.get_trial(trial).status == 'reserved'
            assert client._pacemakers == {}
            remote_pacemaker.stop()

    def test_reserve_race_condition(self):
        """Verify that race conditions during `reserve` is detected and raises a comprehensible
        error
        """
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[0]['_id'])
            experiment.set_trial_status(trial, 'reserved')
            trial.status = 'new'  # Let's pretend it is still available

            with pytest.raises(RuntimeError) as exc:
                client.reserve(trial)

            assert 'Could not reserve trial {}.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}


class TestRelease:
    """Tests for ExperimentClient.release"""

    def test_release(self):
        """Test releasing (to interrupted)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            client.release(trial)
            assert trial.status == 'interrupted'
            assert experiment.get_trial(trial).status == 'interrupted'
            assert trial.id not in client._pacemakers
            assert not pacemaker.is_alive()

    def test_release_status(self):
        """Test releasing with a specific status"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            client.release(trial, 'broken')
            assert trial.status == 'broken'
            assert experiment.get_trial(trial).status == 'broken'
            assert trial.id not in client._pacemakers
            assert not pacemaker.is_alive()

    def test_release_invalid_status(self):
        """Test releasing with a specific status"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            with pytest.raises(ValueError) as exc:
                client.release(trial, 'mouf mouf')

            assert exc.match('Given status `mouf mouf` not one of')

    def test_release_dont_exist(self, monkeypatch):
        """Verify that unregistered trials cannot be released"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment='idontexist', params=cfg.trials[1]['params'])

            def do_nada(trial):
                """Don't do anything"""
                return None
            monkeypatch.setattr(client, '_release_reservation', do_nada)

            with pytest.raises(ValueError) as exc:
                client.release(trial)

            assert 'Trial {} does not exist in database.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}

    def test_release_race_condition(self):
        """Verify that race conditions during `release` is detected and raises a comprehensible
        error
        """
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            # Woops! Trial got failed over from another process.
            experiment.set_trial_status(trial, 'interrupted')
            trial.status = 'reserved'  # Let's pretend we don't know.

            with pytest.raises(RuntimeError) as exc:
                client.release(trial)

            assert ('Reservation for trial {} has been lost before release.'.format(trial.id)
                    in str(exc.value))
            assert client._pacemakers == {}
            assert not pacemaker.is_alive()

    def test_release_unreserved(self):
        """Verify that unreserved trials cannot be released"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]['_id'])
            with pytest.raises(RuntimeError) as exc:
                client.release(trial)

            assert ('Trial {} had no pacemakers. Was is reserved properly?'.format(trial.id) ==
                    str(exc.value))

            assert client._pacemakers == {}


class TestClose:
    """Test close method of the client"""

    def test_close_empty(self):
        """Test client can close when no trial is reserved"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            client.close()

    def test_close_with_reserved(self):
        """Test client cannot be closed if trials are reserved."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()

            with pytest.raises(RuntimeError) as exc:
                client.close()

            assert "There is still reserved trials" in str(exc.value)

            client.release(trial)

    def test_close_unregister_atexit(self, monkeypatch):
        """Test close properly unregister the atexit function"""
        def please_dont_call_me(client):
            raise RuntimeError("Please don't call me!!!")

        monkeypatch.setattr('orion.client.experiment.set_broken_trials', please_dont_call_me)

        with create_experiment(config, base_trial) as (cfg, experiment, client):
            # The registered function in atexit is called as expected
            with pytest.raises(RuntimeError) as exc:
                atexit._run_exitfuncs()

            assert "Please don't call me!!!" == str(exc.value)

            # Unregister the function
            client.close()


class TestBroken:
    """Test handling of broken trials with atexit()"""

    def test_broken_trial(self):
        """Test that broken trials are detected"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'

            atexit._run_exitfuncs()

            assert client._pacemakers == {}
            assert client.get_trial(trial).status == 'broken'

    def test_atexit_with_multiple_clients(self):
        """Test that each client has a separate atexit function"""
        config1 = copy.deepcopy(config)
        config2 = copy.deepcopy(config)
        config2['name'] = 'cloned'
        with create_experiment(exp_config=config1, trial_config=base_trial) as (_, _, client1):
            with create_experiment(exp_config=config2, trial_config=base_trial) as (_, _, client2):
                trial1 = client1.suggest()
                trial2 = client2.suggest()

                assert trial1.status == 'reserved'
                assert trial2.status == 'reserved'

                atexit._run_exitfuncs()

                assert client1._pacemakers == {}
                assert client2._pacemakers == {}
                assert client1.get_trial(trial1).status == 'broken'
                assert client2.get_trial(trial2).status == 'broken'

    def test_atexit_with_multiple_clients_unregister(self, monkeypatch):
        """Test that each client has a separate atexit function that can be unregistered"""
        config1 = copy.deepcopy(config)
        config2 = copy.deepcopy(config)
        config2['name'] = 'cloned'
        with create_experiment(exp_config=config1, trial_config=base_trial) as (_, _, client1):

            def please_dont_call_me(client):
                raise RuntimeError("Please don't call me!!!")

            monkeypatch.setattr('orion.client.experiment.set_broken_trials', please_dont_call_me)

            with create_experiment(exp_config=config2, trial_config=base_trial) as (_, _, client2):
                trial1 = client1.suggest()
                trial2 = client2.suggest()

                # The registered function in atexit is called as expected
                with pytest.raises(RuntimeError) as exc:
                    atexit._run_exitfuncs()

                assert "Please don't call me!!!" == str(exc.value)

                # Unregister the function
                client2.release(trial2)
                client2.close()

                # It should not be called
                atexit._run_exitfuncs()

                assert client1._pacemakers == {}
                assert client2._pacemakers == {}
                assert client1.get_trial(trial1).status == 'broken'
                assert client2.get_trial(trial2).status == 'interrupted'

    def test_interrupted_trial(self):
        """Test that interrupted trials are not set to broken"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'

            try:
                raise KeyboardInterrupt
            except KeyboardInterrupt as e:
                atexit._run_exitfuncs()

            assert client._pacemakers == {}
            assert client.get_trial(trial).status == 'interrupted'


class TestSuggest:
    """Tests for ExperimentClient.suggest"""

    def test_suggest(self):
        """Verify that suggest reserved availabe trials."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'
            assert trial.params['x'] == 1

            assert len(experiment.fetch_trials()) == 5
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_suggest_new(self):
        """Verify that suggest can create, register and reserved new trials."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            for _ in range(3):
                trial = client.suggest()
                assert trial.status == 'reserved'
                assert len(experiment.fetch_trials()) == 5
                assert client._pacemakers[trial.id].is_alive()
                client._pacemakers[trial.id].stop()

            trial = client.suggest()
            assert trial.status == 'reserved'
            assert trial.params['x'] == 57.57
            assert len(experiment.fetch_trials()) == 6

            assert client._pacemakers[trial.id].is_alive()
            for trial_id in list(client._pacemakers.keys()):
                client._pacemakers.pop(trial_id).stop()

    def test_suggest_race_condition(self, monkeypatch):
        """Verify that race conditions to register new trials is handled"""
        new_value = 50.

        # algo will suggest once an already existing trial
        def amnesia(num=1):
            """Suggest a new value and then always suggest the same"""
            if amnesia.count == 0:
                value = [0]
            else:
                value = [new_value]

            amnesia.count += 1

            return [value]

        amnesia.count = 0

        with create_experiment(config, base_trial,
                               statuses=['completed']) as (cfg, experiment, client):

            monkeypatch.setattr(experiment.algorithms, 'suggest', amnesia)

            assert len(experiment.fetch_trials()) == 1

            trial = client.suggest()
            assert trial.status == 'reserved'
            assert trial.params['x'] == new_value
            assert amnesia.count == 2

            assert len(experiment.fetch_trials()) == 2
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_suggest_algo_opt_out(self, monkeypatch):
        """Verify that None is returned when algo cannot sample new trials (opting opt)"""
        def opt_out(num=1):
            """Never suggest a new trial"""
            return None

        monkeypatch.setattr(orion.core.config.worker, 'max_idle_time', 0)

        with create_experiment(config, base_trial,
                               statuses=['completed']) as (cfg, experiment, client):

            monkeypatch.setattr(experiment.algorithms, 'suggest', opt_out)

            assert len(experiment.fetch_trials()) == 1

            with pytest.raises(SampleTimeout):
                client.suggest()

    def test_suggest_is_done(self):
        """Verify that completed experiments cannot suggest new trials"""
        with create_experiment(config, base_trial,
                               statuses=['completed'] * 10) as (cfg, experiment, client):

            assert len(experiment.fetch_trials()) == 10
            assert client.is_done

            assert client.suggest() is None

    def test_suggest_is_broken(self):
        """Verify that broken experiments cannot suggest new trials"""
        with create_experiment(config, base_trial,
                               statuses=['broken'] * 10) as (cfg, experiment, client):

            assert len(experiment.fetch_trials()) == 10
            assert client.is_broken

            with pytest.raises(BrokenExperiment):
                client.suggest()

    def test_suggest_is_done_race_condition(self, monkeypatch):
        """Verify that inability to suggest because is_done becomes True during produce() is
        handled.
        """
        with create_experiment(config, base_trial,
                               statuses=['completed'] * 5) as (cfg, experiment, client):
            def is_done(self):
                """Experiment is done"""
                return True

            def set_is_done():
                """Set is_done while algo is trying to suggest"""
                monkeypatch.setattr(experiment.__class__, 'is_done', property(is_done))

            monkeypatch.setattr(client._producer, 'produce', set_is_done)

            assert len(experiment.fetch_trials()) == 5
            assert not client.is_done

            assert client.suggest() is None

            assert len(experiment.fetch_trials()) == 5
            assert client.is_done

    def test_suggest_is_broken_race_condition(self, monkeypatch):
        """Verify that experiments that gets broken during local algo.suggest gets properly
        handled
        """
        with create_experiment(config, base_trial,
                               statuses=['broken'] * 1) as (cfg, experiment, client):

            def is_broken(self):
                """Experiment is broken"""
                return True

            def set_is_broken():
                """Set is_broken while algo is trying to suggest"""
                monkeypatch.setattr(experiment.__class__, 'is_broken', property(is_broken))

            monkeypatch.setattr(client._producer, 'produce', set_is_broken)

            assert len(experiment.fetch_trials()) == 1
            assert not client.is_broken

            with pytest.raises(BrokenExperiment):
                client.suggest()

            assert len(experiment.fetch_trials()) == 1
            assert client.is_broken

    def test_suggest_hierarchical_space(self):
        """Verify that suggest returns trial with proper hierarchical parameter."""
        exp_config = copy.deepcopy(config)
        exp_config['space'] = {
            'a': {'x': 'uniform(0, 10, discrete=True)'},
            'b': {'y': 'loguniform(1e-08, 1)',
                  'z': 'choices([\'voici\', \'voila\', 2])'}}
        with create_experiment(exp_config=exp_config,
                               trial_config=base_trial,
                               statuses=[]) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'
            assert len(trial.params) == 2
            assert 'x' in trial.params['a']
            assert 'y' in trial.params['b']
            assert 'z' in trial.params['b']

            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()


class TestObserve:
    """Tests for ExperimentClient.observe"""

    def test_observe(self):
        """Verify that `observe()` will update the storage"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            assert trial.results == []
            client.reserve(trial)
            assert get_storage().get_trial(trial).objective is None
            client.observe(trial, [dict(name='objective', type='objective', value=101)])
            assert get_storage().get_trial(trial).objective.value == 101

    def test_observe_unreserved(self):
        """Verify that `observe()` will fail on non-reserved trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            with pytest.raises(RuntimeError) as exc:
                client.observe(trial, [dict(name='objective', type='objective', value=101)])

            assert ('Trial {} had no pacemakers. Was is reserved properly?'.format(trial.id) ==
                    str(exc.value))

    def test_observe_dont_exist(self):
        """Verify that `observe()` will fail on non-registered trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment='idontexist', params=cfg.trials[0]['params'])
            with pytest.raises(ValueError) as exc:
                client.observe(trial, [dict(name='objective', type='objective', value=101)])

            assert 'Trial {} does not exist in database.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}

    def test_observe_bad_results(self):
        """Verify that bad results type is detected and ValueError is raised"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            client.reserve(trial)
            with pytest.raises(ValueError) as exc:
                client.observe(trial, [dict(name='objective', type='bad bad bad', value=101)])

            assert 'Given type, bad bad bad, not one of: ' in str(exc.value)
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_observe_race_condition(self):
        """Verify that race condition during `observe()` is detected and raised"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            experiment.set_trial_status(trial, 'interrupted')
            trial.status = 'reserved'  # Let's pretend it is still reserved

            with pytest.raises(RuntimeError) as exc:
                client.observe(trial, [dict(name='objective', type='objective', value=101)])

            assert 'Reservation for trial {} has been lost.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}


class TestWorkon:
    """Tests for ExperimentClient.workon"""

    def test_workon(self):
        """Verify that workon processes properly"""
        def foo(x):
            return [dict(name='result', type='objective', value=x * 2)]

        with create_experiment(config, base_trial, statuses=[]) as (cfg, experiment, client):
            client.workon(foo, max_trials=5)
            assert len(experiment.fetch_trials()) == 5
            assert client._pacemakers == {}

    def test_workon_partial(self):
        """Verify that partial is properly passed to the function"""
        def foo(x, y):
            return [dict(name='result', type='objective', value=x * 2 + y)]

        with create_experiment(config, base_trial, statuses=[]) as (cfg, experiment, client):
            client.workon(foo, max_trials=10, y=2)
            assert len(experiment.fetch_trials()) == 10
            assert client._pacemakers == {}

    def test_workon_partial_with_override(self):
        """Verify that partial is overriden by trial.params"""
        def foo(x, y):
            return [dict(name='result', type='objective', value=x * 2 + y)]

        ext_config = copy.deepcopy(config)
        ext_config['space']['y'] = 'uniform(0, 10)'

        with create_experiment(exp_config=ext_config,
                               trial_config=base_trial,
                               statuses=[]) as (cfg, experiment, client):
            default_y = 2
            assert len(experiment.fetch_trials()) == 0
            client.workon(foo, max_trials=1, y=default_y)
            assert len(experiment.fetch_trials()) == 1
            assert experiment.fetch_trials()[0].params['y'] != 2

    def test_workon_hierarchical_partial_with_override(self):
        """Verify that hierarchical partial is overriden by trial.params"""
        default_y = 2
        default_z = 'voila'

        def foo(a, b):
            assert b['y'] != default_y
            assert b['z'] == default_z
            return [dict(name='result', type='objective', value=a['x'] * 2 + b['y'])]

        ext_config = copy.deepcopy(config)
        ext_config['space'] = {
            'a': {'x': 'uniform(0, 10, discrete=True)'},
            'b': {'y': 'loguniform(1e-08, 1)'}}

        with create_experiment(exp_config=ext_config,
                               trial_config=base_trial,
                               statuses=[]) as (cfg, experiment, client):
            assert len(experiment.fetch_trials()) == 0
            client.workon(foo, max_trials=5, b={'y': default_y, 'z': default_z})
            assert len(experiment.fetch_trials()) == 5
            params = experiment.fetch_trials()[0].params
            assert len(params)
            assert 'x' in params['a']
            assert 'y' in params['b']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for branching."""

import os

import pytest

import orion.core.cli
import orion.core.io.experiment_builder as experiment_builder
from orion.storage.base import get_storage


def execute(command, assert_code=0):
    """Execute orion command and return returncode"""
    returncode = orion.core.cli.main(command.split(' '))
    assert returncode == assert_code


@pytest.fixture
def init_full_x(clean_db, monkeypatch):
    """Init original experiment"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    name = "full_x"
    orion.core.cli.main(("hunt --init-only -n {name} --config orion_config.yaml ./black_box.py "
                         "-x~uniform(-10,10)").format(name=name).split(" "))
    orion.core.cli.main("insert -n {name} script -x=0".format(name=name).split(" "))


@pytest.fixture
def init_full_x_full_y(init_full_x):
    """Add y dimension to original"""
    name = "full_x"
    branch = "full_x_full_y"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
         "./black_box_with_y.py "
         "-x~uniform(-10,10) "
         "-y~+uniform(-10,10,default_value=1)").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=1 -y=1".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-1 -y=1".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=1 -y=-1".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-1 -y=-1".format(branch=branch).split(" "))


@pytest.fixture
def init_half_x_full_y(init_full_x_full_y):
    """Change x's prior to full x and full y experiment"""
    name = "full_x_full_y"
    branch = "half_x_full_y"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
         "-x~+uniform(0,10) "
         "-y~uniform(-10,10,default_value=1)").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=2 -y=2".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=2 -y=-2".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_half_y(init_full_x_full_y):
    """Change y's prior to full x and full y experiment"""
    name = "full_x_full_y"
    branch = "full_x_half_y"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
         "-x~uniform(-10,10) "
         "-y~+uniform(0,10,default_value=1)").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=3 -y=3".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-3 -y=3".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_rename_y_z(init_full_x_full_y):
    """Rename y from full x full y to z"""
    name = "full_x_full_y"
    branch = "full_x_rename_y_z"
    orion.core.cli.main((
        "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
        "./black_box_with_z.py -x~uniform(-10,10) -y~>z -z~uniform(-10,10,default_value=1)"
        ).format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=4 -z=4".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-4 -z=4".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=4 -z=-4".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-4 -z=-4".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_rename_half_y_half_z(init_full_x_half_y):
    """Rename y from full x half y to z"""
    name = "full_x_half_y"
    branch = "full_x_rename_half_y_half_z"
    orion.core.cli.main((
        "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
        "./black_box_with_z.py -x~uniform(-10,10) -y~>z -z~uniform(0,10,default_value=1)"
        ).format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=5 -z=5".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-5 -z=5".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_rename_half_y_full_z(init_full_x_half_y):
    """Rename y from full x half y to full z (rename + changed prior)"""
    name = "full_x_half_y"
    branch = "full_x_rename_half_y_full_z"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
         "./black_box_with_z.py "
         "-x~uniform(-10,10) -y~>z "
         "-z~+uniform(-10,10,default_value=1)").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=6 -z=6".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-6 -z=6".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=6 -z=-6".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-6 -z=-6".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_remove_y(init_full_x_full_y):
    """Remove y from full x full y"""
    name = "full_x_full_y"
    branch = "full_x_remove_y"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
         "./black_box.py "
         "-x~uniform(-10,10) -y~-").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=7".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-7".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_remove_z(init_full_x_rename_y_z):
    """Remove z from full x full z"""
    name = "full_x_rename_y_z"
    branch = "full_x_remove_z"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
         "./black_box.py "
         "-x~uniform(-10,10) -z~-").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=8".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-8".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_remove_z_default_4(init_full_x_rename_y_z):
    """Remove z from full x full z and give a default value of 4"""
    name = "full_x_rename_y_z"
    branch = "full_x_remove_z_default_4"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
         "./black_box.py "
         "-x~uniform(-10,10) -z~-4").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=9".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-9".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_new_algo(init_full_x):
    """Remove z from full x full z and give a default value of 4"""
    name = "full_x"
    branch = "full_x_new_algo"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} "
         "--algorithm-change --config new_algo_config.yaml "
         "./black_box.py -x~uniform(-10,10)").format(name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=1.1".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-1.1".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_new_cli(init_full_x):
    """Remove z from full x full z and give a default value of 4"""
    name = "full_x"
    branch = "full_x_new_cli"
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
         "./black_box_new.py -x~uniform(-10,10) --a-new argument").format(
             name=name, branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=1.2".format(branch=branch).split(" "))
    orion.core.cli.main("insert -n {branch} script -x=-1.2".format(branch=branch).split(" "))


@pytest.fixture
def init_full_x_ignore_cli(init_full_x):
    """Use the --non-monitored-arguments argument"""
    name = "full_x_with_new_opt"
    orion.core.cli.main(("hunt --init-only -n {name} --config orion_config.yaml ./black_box_new.py "
                         "-x~uniform(-10,10)").format(name=name).split(" "))
    orion.core.cli.main("insert -n {name} script -x=0".format(name=name).split(" "))

    orion.core.cli.main(
        ("hunt --init-only -n {name} --non-monitored-arguments a-new "
         "--config orion_config.yaml ./black_box_new.py "
         "-x~uniform(-10,10) --a-new argument").format(name=name).split(" "))
    orion.core.cli.main("insert -n {name} script -x=1.2".format(name=name).split(" "))
    orion.core.cli.main("insert -n {name} script -x=-1.2".format(name=name).split(" "))


@pytest.fixture
def init_entire(init_half_x_full_y,  # 1.1.1
                init_full_x_rename_half_y_half_z,  # 1.1.2.1
                init_full_x_rename_half_y_full_z,  # 1.1.2.2
                init_full_x_remove_y,  # 1.1.4
                init_full_x_remove_z,  # 1.1.3.1
                init_full_x_remove_z_default_4):  # 1.1.3.2
    """Initialize all experiments"""
    pass


def get_name_value_pairs(trials):
    """Turn parameters into pairs for easy comparisions"""
    pairs = []
    for trial in trials:
        pairs.append([])
        for param in trial._params:
            pairs[-1].append((param.name, param.value))

        pairs[-1] = tuple(pairs[-1])

    return tuple(pairs)


def test_init(init_full_x, create_db_instance):
    """Test if original experiment contains trial 0"""
    experiment = experiment_builder.build_view(name='full_x')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 0), ), )


def test_full_x_full_y(init_full_x_full_y, create_db_instance):
    """Test if full x full y is properly initialized and can fetch original trial"""
    experiment = experiment_builder.build_view(name='full_x_full_y')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 1), ('/y', 1)),
                     (('/x', -1), ('/y', 1)),
                     (('/x', 1), ('/y', -1)),
                     (('/x', -1), ('/y', -1)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == ((('/x', 0), ('/y', 1)),
                     (('/x', 1), ('/y', 1)),
                     (('/x', -1), ('/y', 1)),
                     (('/x', 1), ('/y', -1)),
                     (('/x', -1), ('/y', -1)))


def test_half_x_full_y(init_half_x_full_y, create_db_instance):
    """Test if half x full y is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.build_view(name='half_x_full_y')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 2), ('/y', 2)),
                     (('/x', 2), ('/y', -2)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == ((('/x', 0), ('/y', 1)),
                     (('/x', 1), ('/y', 1)),
                     (('/x', 1), ('/y', -1)),
                     (('/x', 2), ('/y', 2)),
                     (('/x', 2), ('/y', -2)))


def test_full_x_half_y(init_full_x_half_y, create_db_instance):
    """Test if full x half y is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.build_view(name='full_x_half_y')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 3), ('/y', 3)),
                     (('/x', -3), ('/y', 3)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == ((('/x', 0), ('/y', 1)),
                     (('/x', 1), ('/y', 1)),
                     (('/x', -1), ('/y', 1)),
                     (('/x', 3), ('/y', 3)),
                     (('/x', -3), ('/y', 3)))


def test_full_x_rename_y_z(init_full_x_rename_y_z, create_db_instance):
    """Test if full x full z is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.build_view(name='full_x_rename_y_z')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 4), ('/z', 4)),
                     (('/x', -4), ('/z', 4)),
                     (('/x', 4), ('/z', -4)),
                     (('/x', -4), ('/z', -4)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == ((('/x', 0), ('/z', 1)),
                     (('/x', 1), ('/z', 1)),
                     (('/x', -1), ('/z', 1)),
                     (('/x', 1), ('/z', -1)),
                     (('/x', -1), ('/z', -1)),
                     (('/x', 4), ('/z', 4)),
                     (('/x', -4), ('/z', 4)),
                     (('/x', 4), ('/z', -4)),
                     (('/x', -4), ('/z', -4)))


def test_full_x_rename_half_y_half_z(init_full_x_rename_half_y_half_z, create_db_instance):
    """Test if full x half z is properly initialized and can fetch from its 3 parents"""
    experiment = experiment_builder.build_view(name='full_x_rename_half_y_half_z')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 5), ('/z', 5)),
                     (('/x', -5), ('/z', 5)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == ((('/x', 0), ('/z', 1)),
                     (('/x', 1), ('/z', 1)),
                     (('/x', -1), ('/z', 1)),
                     (('/x', 3), ('/z', 3)),
                     (('/x', -3), ('/z', 3)),
                     (('/x', 5), ('/z', 5)),
                     (('/x', -5), ('/z', 5)))


def test_full_x_rename_half_y_full_z(init_full_x_rename_half_y_full_z, create_db_instance):
    """Test if full x half->full z is properly initialized and can fetch from its 3 parents"""
    experiment = experiment_builder.build_view(name='full_x_rename_half_y_full_z')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 6), ('/z', 6)),
                     (('/x', -6), ('/z', 6)),
                     (('/x', 6), ('/z', -6)),
                     (('/x', -6), ('/z', -6)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == ((('/x', 0), ('/z', 1)),
                     (('/x', 1), ('/z', 1)),
                     (('/x', -1), ('/z', 1)),
                     (('/x', 3), ('/z', 3)),
                     (('/x', -3), ('/z', 3)),
                     (('/x', 6), ('/z', 6)),
                     (('/x', -6), ('/z', 6)),
                     (('/x', 6), ('/z', -6)),
                     (('/x', -6), ('/z', -6)))


def test_full_x_remove_y(init_full_x_remove_y, create_db_instance):
    """Test if full x removed y is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.build_view(name='full_x_remove_y')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 7), ), (('/x', -7), ))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == ((('/x', 0), ), (('/x', 1), ), (('/x', -1), ), (('/x', 7), ), (('/x', -7), ))


def test_full_x_remove_z(init_full_x_remove_z, create_db_instance):
    """Test if full x removed z is properly initialized and can fetch from 2 of its 3 parents"""
    experiment = experiment_builder.build_view(name='full_x_remove_z')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 8), ), (('/x', -8), ))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    # Note that full_x_rename_y_z are filtered out because default_value=1
    assert pairs == ((('/x', 0), ), (('/x', 1), ), (('/x', -1), ), (('/x', 8), ), (('/x', -8), ))


def test_full_x_remove_z_default_4(init_full_x_remove_z_default_4, create_db_instance):
    """Test if full x removed z  (default 4) is properly initialized and can fetch
    from 1 of its 3 parents
    """
    experiment = experiment_builder.build_view(name='full_x_remove_z_default_4')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 9), ), (('/x', -9), ))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    # Note that full_x and full_x_full_y are filtered out because default_value=4
    assert pairs == ((('/x', 4), ), (('/x', -4), ), (('/x', 9), ), (('/x', -9), ))


def test_entire_full_x_full_y(init_entire, create_db_instance):
    """Test if full x full y can fetch from its parent and all children"""
    experiment = experiment_builder.build_view(name='full_x_full_y')
    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((('/x', 1), ('/y', 1)),
                     (('/x', -1), ('/y', 1)),
                     (('/x', 1), ('/y', -1)),
                     (('/x', -1), ('/y', -1)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert set(pairs) == set(((('/x', 0), ('/y', 1)),
                              # full_x_full_y
                              (('/x', 1), ('/y', 1)),
                              (('/x', -1), ('/y', 1)),
                              (('/x', 1), ('/y', -1)),
                              (('/x', -1), ('/y', -1)),
                              # half_x_full_y
                              (('/x', 2), ('/y', 2)),
                              (('/x', 2), ('/y', -2)),
                              # full_x_half_y
                              (('/x', 3), ('/y', 3)),
                              (('/x', -3), ('/y', 3)),
                              # full_x_rename_y_z
                              (('/x', 4), ('/y', 4)),
                              (('/x', -4), ('/y', 4)),
                              (('/x', 4), ('/y', -4)),
                              (('/x', -4), ('/y', -4)),
                              # full_x_rename_half_y_half_z
                              (('/x', 5), ('/y', 5)),
                              (('/x', -5), ('/y', 5)),
                              # full_x_rename_half_y_full_z
                              (('/x', 6), ('/y', 6)),
                              (('/x', -6), ('/y', 6)),
                              # full_x_remove_y
                              (('/x', 7), ('/y', 1)),
                              (('/x', -7), ('/y', 1)),
                              # full_x_remove_z
                              (('/x', 8), ('/y', 1)),
                              (('/x', -8), ('/y', 1)),
                              # full_x_remove_z_default_4
                              (('/x', 9), ('/y', 4)),
                              (('/x', -9), ('/y', 4))))


def test_run_entire_full_x_full_y(init_entire, create_db_instance):
    """Test if branched experiment can be executed without triggering a branching event again"""
    experiment = experiment_builder.build_view(name='full_x_full_y')
    assert len(experiment.fetch_trials(with_evc_tree=True)) == 23
    assert len(experiment.fetch_trials()) == 4

    orion.core.cli.main(("-vv hunt --max-trials 20 --pool-size 1 -n full_x_full_y "
                         "./black_box_with_y.py "
                         "-x~uniform(-10,10) "
                         "-y~uniform(-10,10,default_value=1)").split(" "))

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 39
    assert len(experiment.fetch_trials()) == 20


def test_run_entire_full_x_full_y_no_args(init_entire, create_db_instance):
    """Test if branched experiment can be executed without script arguments"""
    experiment = experiment_builder.build_view(name='full_x_full_y')
    assert len(experiment.fetch_trials(with_evc_tree=True)) == 23
    assert len(experiment.fetch_trials()) == 4

    orion.core.cli.main(("-vv hunt --max-trials 20 --pool-size 1 -n full_x_full_y").split(" "))

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 39
    assert len(experiment.fetch_trials()) == 20


def test_new_algo(init_full_x_new_algo):
    """Test that new algo conflict is automatically resolved"""
    experiment = experiment_builder.build_view(name='full_x_new_algo')
    assert len(experiment.fetch_trials(with_evc_tree=True)) == 3
    assert len(experiment.fetch_trials()) == 2

    orion.core.cli.main(("-vv hunt --max-trials 20 --pool-size 1 -n full_x_new_algo").split(" "))

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 21
    assert len(experiment.fetch_trials()) == 20


def test_new_algo_not_resolved(init_full_x, capsys):
    """Test that new algo conflict is not automatically resolved"""
    name = "full_x"
    branch = "full_x_new_algo"
    error_code = orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} --config new_algo_config.yaml "
         "--manual-resolution ./black_box.py -x~uniform(-10,10)")
        .format(name=name, branch=branch).split(" "))
    assert error_code == 1

    captured = capsys.readouterr()
    assert captured.out == ''
    assert "Configuration is different and generates a branching event" in captured.err
    assert "gradient_descent" in captured.err


def test_ignore_cli(init_full_x_ignore_cli):
    """Test that a non-monitored parameter conflict is not generating a child"""
    name = "full_x"
    orion.core.cli.main(
        ("hunt --init-only -n {name} --non-monitored-arguments a-new "
         "--manual-resolution ./black_box.py -x~uniform(-10,10)")
        .format(name=name).split(" "))


@pytest.mark.usefixtures('init_full_x', 'mock_infer_versioning_metadata')
def test_new_code_triggers_code_conflict(capsys):
    """Test that a different git hash is generating a child"""
    name = "full_x"
    error_code = orion.core.cli.main(
        ("hunt --init-only -n {name} "
         "--manual-resolution ./black_box.py -x~uniform(-10,10)")
        .format(name=name).split(" "))
    assert error_code == 1

    captured = capsys.readouterr()
    assert captured.out == ''
    assert "Configuration is different and generates a branching event" in captured.err
    assert "--code-change-type" in captured.err


@pytest.mark.usefixtures('init_full_x', 'mock_infer_versioning_metadata')
def test_new_code_ignores_code_conflict():
    """Test that a different git hash is *not* generating a child if --ignore-code-changes"""
    name = "full_x"
    orion.core.cli.main(
        ("hunt --init-only -n {name} --ignore-code-changes "
         "--manual-resolution ./black_box.py -x~uniform(-10,10)")
        .format(name=name).split(" "))


def test_new_cli(init_full_x_new_cli):
    """Test that new cli conflict is automatically resolved"""
    experiment = experiment_builder.build_view(name='full_x_new_cli')
    assert len(experiment.fetch_trials(with_evc_tree=True)) == 3
    assert len(experiment.fetch_trials()) == 2

    orion.core.cli.main(("-vv hunt --max-trials 20 --pool-size 1 -n full_x_new_cli").split(" "))

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 21
    assert len(experiment.fetch_trials()) == 20


def test_auto_resolution_does_resolve(init_full_x_full_y, monkeypatch):
    """Test that auto-resolution does resolve all conflicts"""
    # Patch cmdloop to avoid autoresolution's prompt
    monkeypatch.setattr('sys.__stdin__.isatty', lambda: True)

    name = "full_x_full_y"
    branch = "half_x_no_y_new_w"
    # If autoresolution was not succesfull, this to fail with a sys.exit without registering the
    # experiment
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
         "-x~uniform(0,10) "
         "-w~choices(['a','b'])").format(name=name, branch=branch).split(" "))


def test_auto_resolution_with_fidelity(init_full_x_full_y, monkeypatch):
    """Test that auto-resolution does resolve all conflicts including new fidelity"""
    # Patch cmdloop to avoid autoresolution's prompt
    monkeypatch.setattr('sys.__stdin__.isatty', lambda: True)

    name = "full_x_full_y"
    branch = "half_x_no_y_new_w"
    # If autoresolution was not succesfull, this to fail with a sys.exit without registering the
    # experiment
    orion.core.cli.main(
        ("hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
         "-x~uniform(0,10, precision=None) "
         "-w~fidelity(1,10)").format(name=name, branch=branch).split(" "))


def test_init_w_version_from_parent_w_children(clean_db, monkeypatch, capsys):
    """Test that init of experiment from version with children fails."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute("hunt --init-only -n experiment --config orion_config.yaml "
            "./black_box.py -x~normal(0,1)")
    execute("hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)")

    execute(
        "hunt --init-only -n experiment -v 1 "
        "./black_box.py -x~normal(0,1) -y~+normal(0,1) -z~normal(0,1)",
        assert_code=1)

    captured = capsys.readouterr()
    assert captured.out == ''
    assert "Configuration is different and generates a branching event" in captured.err
    assert "Experiment name" in captured.err


def test_init_w_version_from_exp_wout_child(clean_db, monkeypatch):
    """Test that init of experiment from version without child works."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute("hunt --init-only -n experiment --config orion_config.yaml "
            "./black_box.py -x~normal(0,1)")
    execute("hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)")
    execute("hunt --init-only -n experiment -v 2 ./black_box.py "
            "-x~normal(0,1) -y~+normal(0,1) -z~+normal(0,1)")

    exp = get_storage().fetch_experiments({'name': 'experiment', 'version': 3})
    assert len(list(exp))


def test_init_w_version_gt_max(clean_db, monkeypatch):
    """Test that init of experiment from version higher than max works."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute("hunt --init-only -n experiment --config orion_config.yaml "
            "./black_box.py -x~normal(0,1)")
    execute("hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)")
    execute("hunt --init-only -n experiment -v 2000 ./black_box.py "
            "-x~normal(0,1) -y~+normal(0,1) -z~+normal(0,1)")

    exp = get_storage().fetch_experiments({'name': 'experiment', 'version': 3})
    assert len(list(exp))


def test_init_check_increment_w_children(clean_db, monkeypatch):
    """Test that incrementing version works with not same-named children."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute("hunt --init-only -n experiment --config orion_config.yaml "
            "./black_box.py -x~normal(0,1)")
    execute("hunt --init-only -n experiment --branch-to experiment_2 ./black_box.py "
            "-x~normal(0,1) -y~+normal(0,1)")
    execute("hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -z~+normal(0,1)")

    exp = get_storage().fetch_experiments({'name': 'experiment', 'version': 2})
    assert len(list(exp))


def test_branch_from_selected_version(clean_db, monkeypatch):
    """Test that branching from a version passed with `--version` works."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute("hunt --init-only -n experiment --config orion_config.yaml "
            "./black_box.py -x~normal(0,1)")
    execute("hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)")
    execute("hunt --init-only -n experiment --version 1 -b experiment_2 ./black_box.py "
            "-x~normal(0,1) -z~+normal(0,1)")

    storage = get_storage()
    parent = storage.fetch_experiments({'name': 'experiment', 'version': 1})[0]
    exp = storage.fetch_experiments({'name': 'experiment_2'})[0]
    assert exp['refers']['parent_id'] == parent['_id']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.info` -- Module to info experiments
========================================================

.. module:: info
   :platform: Unix
   :synopsis: Commandline support to print details of experiments in terminal

"""
import logging
import sys

from orion.core.cli.base import get_basic_args_group
import orion.core.io.experiment_builder as experiment_builder
from orion.core.utils.format_terminal import format_info

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = 'Gives detailed information about experiments'


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    info_parser = parser.add_parser('info', help=SHORT_DESCRIPTION, description=SHORT_DESCRIPTION)
    get_basic_args_group(info_parser)

    info_parser.set_defaults(func=main)

    return info_parser


def main(args):
    """Fetch config and info experiments"""
    try:
        experiment = experiment_builder.build_view_from_args(args)
    except ValueError:
        print('Experiment {} not found in db.'.format(args.get('name', None)))
        sys.exit(1)

    print(format_info(experiment))



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: disable-all
"""
:mod:`orion.core.utils.pptree` -- Utilitary functions for printing trees
========================================================================

.. module:: pptree
   :platform: Unix
   :synopsis: Utilitary functions to provided pretty trees


Clement Michard (c) 2015

https://github.com/clemtoy/pptree

MIT License

Copyright (c) 2017 Clément Michard

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

class Node:
    def __init__(self, name, parent=None):
        self.name = name
        self.parent = parent
        self.children = []

        if parent:
            self.parent.children.append(self)


def print_tree(current_node, childattr='children', nameattr='name', indent='', last='updown'):
    if hasattr(current_node, nameattr):
        name = lambda node: getattr(node, nameattr)
    else:
        name = lambda node: str(node)

    children = lambda node: getattr(node, childattr)
    nb_children = lambda node: sum(nb_children(child) for child in children(node)) + 1
    size_branch = {child: nb_children(child) for child in children(current_node)}

    """ Creation of balanced lists for "up" branch and "down" branch. """
    up = sorted(children(current_node), key=lambda node: nb_children(node))
    down = []
    while up and sum(size_branch[node] for node in down) < sum(size_branch[node] for node in up):
        down.append(up.pop())

    """ Printing of "up" branch. """
    for child in up:     
        next_last = 'up' if up.index(child) is 0 else ''
        next_indent = '{0}{1}{2}'.format(indent, ' ' if 'up' in last else '│', ' ' * len(name(current_node)))
        print_tree(child, childattr, nameattr, next_indent, next_last)

    """ Printing of current node. """
    if last == 'up': start_shape = '┌'
    elif last == 'down': start_shape = '└'
    elif last == 'updown': start_shape = ' '
    else: start_shape = '├'

    if up: end_shape = '┤'
    elif down: end_shape = '┐'
    else: end_shape = ''

    print('{0}{1}{2}{3}'.format(indent, start_shape, name(current_node), end_shape))

    """ Printing of "down" branch. """
    for child in down:
        next_last = 'down' if down.index(child) is len(down) - 1 else ''
        next_indent = '{0}{1}{2}'.format(indent, ' ' if 'down' in last else '│', ' ' * len(name(current_node)))
        print_tree(child, childattr, nameattr, next_indent, next_last)
